from typing import cast

import bluish.actions.base
import bluish.nodes.job
import bluish.nodes.step
import bluish.process
from bluish.logging import debug, error, info, warning
from bluish.schemas import Bool, DefaultStringDict, List, Object, Optional, Str
from bluish.utils import decorate_for_log


def _build_list_opt(opt: str, items: list[str] | None) -> str:
    if not items:
        return ""
    if isinstance(items, str):
        items = [items]
    result = " ".join(f"{opt} {o}" for o in items)
    return f" {result}"


def _build_opt(opt: str, value: str | None) -> str:
    if not value:
        return ""
    return f" {opt} {value}"


def _build_flag(flag: str, value: bool | None) -> str:
    return "" if not value else f" {flag}"


def _is_valid_docker_id(id: str) -> bool:
    return len(id) in (12, 64) and all(c in "0123456789abcdef" for c in id)


def run_and_get_pid(command: str, step: bluish.nodes.step.Step) -> str:
    job = cast(bluish.nodes.job.Job, step.parent)
    result = job.exec(command, step)
    return result.stdout.strip() if result.returncode == 0 else ""


def docker_ps(
    step: bluish.nodes.step.Step,
    name: str | None = None,
    pid: str | None = None,
) -> bluish.process.ProcessResult:
    filter = f"name={name}" if name else f"id={pid}"
    job = cast(bluish.nodes.job.Job, step.parent)
    return job.exec(f"docker ps -f {filter} --all --quiet", step)


class Login(bluish.actions.base.Action):
    FQN: str = "docker/login"

    INPUTS_SCHEMA = Object(
        {
            "username": Str,
            "password": Str,
            "registry": Optional(Str),
        }
    )

    SENSITIVE_INPUTS: tuple[str, ...] = ("password",)

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        inputs = step.inputs

        username = inputs["username"]
        password = inputs["password"]
        registry = inputs.get("registry", "")

        command = step.expand_expr(
            f"docker login --username '{username}' --password '{password}' {registry}"
        )
        if step.get_inherited_attr("echo_commands", True):
            protected_command = (
                f"docker login --username '{username}' --password ******** {registry}"
            )
            info(f"Docker login:\n -> {protected_command}")

        job = cast(bluish.nodes.job.Job, step.parent)
        login_result = job.exec(command, step, stream_output=True)
        if login_result.failed:
            error(f"Login failed: {login_result.error}")
        return login_result


class Logout(bluish.actions.base.Action):
    FQN: str = "docker/logout"

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        command = "docker logout"
        if step.get_inherited_attr("echo_commands", True):
            info("Logging out of Docker...")
        job = cast(bluish.nodes.job.Job, step.parent)
        return job.exec(command, step)


class Build(bluish.actions.base.Action):
    FQN: str = "docker/build"

    INPUTS_SCHEMA = Object(
        {
            "dockerfile": Optional(Str),
            "tags": Optional(Str, List(Str)),
            "context": Optional(Str),
        }
    )

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        inputs = step.inputs

        dockerfile = inputs.get("dockerfile", "Dockerfile")

        options = f"-f '{dockerfile}'"
        options += _build_list_opt("-t", inputs["tags"])
        working_dir = step.get_inherited_attr("working_directory", ".")
        context = inputs.get("context", working_dir)
        command = step.expand_expr(f"docker build {options} {context}")
        if step.get_inherited_attr("echo_commands", True):
            info(f"Building image:\n -> {command}")

        job = cast(bluish.nodes.job.Job, step.parent)
        build_result = job.exec(command, step, stream_output=True)
        if build_result.failed:
            error(f"Failed to build image: {build_result.error}")
        return build_result


class GetPid(bluish.actions.base.Action):
    FQN: str = "docker/get-pid"

    INPUTS_SCHEMA = Object(
        {
            "name": Str,
        }
    )

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        name = step.inputs["name"]

        ps_result = docker_ps(step, name=name)
        pid = ps_result.stdout.strip()
        if ps_result.failed or not _is_valid_docker_id(pid):
            error(f"Failed to get container id for {name}: {ps_result.error}")
            return (
                ps_result
                if ps_result.failed
                else bluish.process.ProcessResult(
                    returncode=1, stdout=ps_result.stdout, stderr=ps_result.stderr
                )
            )

        return bluish.process.ProcessResult(
            stdout=docker_ps(step, name=name).stdout.strip()
        )


class Run(bluish.actions.base.Action):
    FQN: str = "docker/run"

    INPUTS_SCHEMA = Object(
        {
            "image": Str,
            "name": Str,
            "fail_if_running": Optional(bool),
            "ports": Optional(Str, List(Str)),
            "volumes": Optional(Str, List(Str)),
            "env": Optional(DefaultStringDict),
            "env_file": Optional(Str),
            "network": Optional(Str),
            "label": Optional(Str, List(Str)),
            "pull": Optional(Str),
            "user": Optional(Str),
            "remove": Optional(bool),
            "quiet": Optional(bool),
        }
    )

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        inputs = step.inputs

        image = inputs["image"]
        name = inputs["name"]

        info(f"Running container with image {image} and name {name}...")
        ps_result = docker_ps(step, name=name)
        if ps_result.failed:
            error(
                f"Failed to check if container with name {name} is already running: {ps_result.error}"
            )
            return ps_result

        container_pid = ps_result.stdout.strip()
        if _is_valid_docker_id(container_pid):
            msg = f"Container with name {name} is already running with id {container_pid}."
            if inputs.get("fail_if_running", True):
                error(msg)
                return bluish.process.ProcessResult(returncode=1, stdout=container_pid)
            else:
                warning(msg)
                return bluish.process.ProcessResult(stdout=container_pid)

        options = f"--name {name} --detach"
        options += _build_list_opt("-p", inputs.get("ports"))
        options += _build_list_opt("-v", inputs.get("volumes"))
        options += _build_list_opt("-e", inputs.get("env"))
        options += _build_list_opt("--env-file", inputs.get("env_file"))
        options += _build_flag("--rm", inputs.get("remove"))

        for opt in ["network", "label", "pull", "user"]:
            options += _build_opt(f"--{opt}", inputs.get(opt))
        for flag in ["quiet"]:
            options += _build_flag(f"--{flag}", inputs.get(flag))

        job = cast(bluish.nodes.job.Job, step.parent)
        run_result = job.exec(f"docker run {options} {image}", step)
        if run_result.failed:
            error(f"Failed to start container with image {image}: {run_result.error}")
            return run_result

        container_pid = run_result.stdout.strip()
        if not _is_valid_docker_id(container_pid):
            error(f"Failed to get container id for {name}: {container_pid}")
            return bluish.process.ProcessResult(
                returncode=1, stdout=run_result.stdout, stderr=run_result.stderr
            )

        info(f"Container started with id {container_pid}.")
        return bluish.process.ProcessResult(stdout=container_pid)


class Stop(bluish.actions.base.Action):
    FQN: str = "docker/stop"

    INPUTS_SCHEMA = Object(
        {
            "name": Optional(Str),
            "pid": Optional(Str),
            "signal": Optional(Str),
            "time": Optional(Str),
            "remove": Optional(bool),
            "fail_if_not_found": Optional(bool),
        }
    )

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        inputs = step.inputs

        name = inputs.get("name")
        container_pid = inputs.get("pid")
        input_attr = f"name {name}" if name else f"pid {container_pid}"
        remove_container = inputs.get("remove", False)
        stop_container = True

        job = cast(bluish.nodes.job.Job, step.parent)

        info(f"Stopping container with {input_attr}...")

        ps_result = docker_ps(step, name=name, pid=container_pid)
        if ps_result.failed:
            msg = f"Can't find a container with {input_attr}."
            if inputs.get("fail_if_not_found", True):
                error(msg)
                return ps_result
            else:
                warning(msg)
                stop_container = False
                # If don't need to remove the container, we can stop here
                if not remove_container:
                    return bluish.process.ProcessResult()

        container_pid = ps_result.stdout.strip()
        if not _is_valid_docker_id(container_pid):
            error(
                f"Failed to verify container id for container with {input_attr}: {container_pid}"
            )
            return bluish.process.ProcessResult(
                returncode=1, stdout=ps_result.stdout, stderr=ps_result.stderr
            )

        if name:
            info(f"Container found with id {container_pid}.")

        if stop_container:
            info(f"Stopping container with {input_attr}...")
            options = ""
            for opt in ["signal", "time"]:
                options += _build_opt(f"--{opt}", inputs.get(opt))

            stop_result = job.exec(
                f"docker container stop {options} {container_pid}", step
            )
            if stop_result.failed:
                error(
                    f"Failed to stop container with {input_attr}: {stop_result.error}"
                )
                return stop_result

        if remove_container:
            rm_result = job.exec(f"docker container rm {options} {container_pid}", step)
            if rm_result.failed:
                error(
                    f"Failed to remove container with {input_attr}: {rm_result.error}"
                )
                return rm_result

        return bluish.process.ProcessResult(stdout=container_pid)


class Exec(bluish.actions.base.Action):
    FQN: str = "docker/exec"

    INPUTS_SCHEMA = Object(
        {
            "name": Optional(Str),
            "pid": Optional(Str),
            "run": Str,
            "env": Optional(DefaultStringDict),
            "env_file": Optional(Str),
            "workdir": Optional(Str),
        }
    )

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        inputs = step.inputs

        name = inputs.get("name")
        container_pid = inputs.get("pid")
        command = inputs["run"]
        input_attr = f"name {name}" if name else f"pid {container_pid}"

        pid_result = docker_ps(step, name=name, pid=container_pid)
        if pid_result.failed:
            error(
                f"Can't find a running container with {input_attr}: {pid_result.error}"
            )
            return (
                pid_result
                if pid_result.failed
                else bluish.process.ProcessResult(returncode=1)
            )

        container_pid = pid_result.stdout.strip()
        if not _is_valid_docker_id(container_pid):
            error(
                f"Failed to verify container id for container with {input_attr}: {container_pid}"
            )
            return bluish.process.ProcessResult(
                returncode=1, stdout=pid_result.stdout, stderr=pid_result.stderr
            )

        options = ""
        options += _build_list_opt("-e", inputs.get("env"))
        options += _build_list_opt("--env-file", inputs.get("env_file"))

        for opt in ["workdir"]:
            options += _build_opt(f"--{opt}", inputs.get(opt))
        output = ""

        echo_commands = step.get_inherited_attr("echo_commands", True)
        echo_output = step.get_inherited_attr("echo_output", False)

        i = 0
        command_lines = command.splitlines()
        while i < len(command_lines):
            line = command_lines[i].strip()
            while line.endswith("\\"):
                i += 1
                line = line[:-1] + command_lines[i].strip()
            i += 1

            if echo_commands:
                info(line)

            job = cast(bluish.nodes.job.Job, step.parent)
            result = job.exec(f"docker exec {options} {container_pid} {line}", step)
            output += result.stdout

            if echo_output:
                info(decorate_for_log(result.stdout, "  > "))

            if result.failed:
                if echo_output:
                    error(decorate_for_log(result.error, " ** "))
                return bluish.process.ProcessResult(
                    returncode=result.returncode, stdout=output, stderr=result.stderr
                )

        return bluish.process.ProcessResult(stdout=output)


class CreateNetwork(bluish.actions.base.Action):
    FQN: str = "docker/create-network"

    INPUTS_SCHEMA = Object(
        {
            "name": Str,
            "fail_if_exists": Optional(bool),
            "label": Optional(Str, List(Str)),
            "ingress": Optional(Bool),
            "internal": Optional(Bool),
        }
    )

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        inputs = step.inputs

        name = inputs["name"]
        job = cast(bluish.nodes.job.Job, step.parent)

        info(f"Creating network {name}...")

        debug(f"Checking if network {name} already exists...")
        network_ls_result = job.exec(f"docker network ls -f name={name} --quiet", step)
        if network_ls_result.failed:
            error(f"Failed to list networks: {network_ls_result.error}")
            return network_ls_result

        network_id = network_ls_result.stdout.strip()
        if network_id:
            msg = f"Network {name} already exists with id {network_id}."
            if inputs.get("fail_if_exists", True):
                error(msg)
                return bluish.process.ProcessResult(returncode=1)
            else:
                warning(msg)
        else:
            options = "--attachable"
            for opt in ["label"]:
                options += _build_opt(f"--{opt}", inputs.get(opt))
            for flag in ["ingress", "internal"]:
                options += _build_flag(f"--{flag}", inputs.get(flag))

            network_create_result = job.exec(
                f"docker network create {options} {name}", step
            )
            if network_create_result.failed:
                error(f"Failed to create network {name}: {network_create_result.error}")
                return network_create_result

            network_id = network_create_result.stdout.strip()
            if not _is_valid_docker_id(network_id):
                error(f"Failed to get network id for {name}: {network_id}")
                return bluish.process.ProcessResult(
                    returncode=1,
                    stdout=network_create_result.stdout,
                    stderr=network_create_result.stderr,
                )
            info(f"Network {name} created with id {network_id}.")

        return bluish.process.ProcessResult(stdout=network_id)




from io import FileIO
from test.utils import create_workflow

import pytest
from bluish.core import init_commands


@pytest.fixture(scope="session", autouse=True)
def initialize_commands():
    init_commands()


#@pytest.mark.docker
def test_docker_build(temp_file: FileIO) -> None:
    filename = str(temp_file.name)

    wf = create_workflow(f"""
jobs:
    create-docker:
        steps:
            - name: "Build Alpine Dockerfile"
              uses: core/expand-template
              with:
                  input: |
                      FROM alpine:latest
                      RUN apk add --no-cache python3 py-pip
                      RUN pip3 install bluish==0.0.30 --break-system-packages
                  output_file: {filename}
            - uses: docker/build
              with:
                  dockerfile: {filename}
                  context: .
                  tags:
                      - "bluish-test-alpine:0.0.30"
            - run: |
                  echo "id=$(docker image ls -f reference=bluish-test-alpine:0.0.30 -q)" >> "$BLUISH_OUTPUT"
              set:
                  workflow.var.docker-image-id: ${{{{ outputs.id }}}}
            - run: |
                  docker image rm $(docker image ls -f reference=bluish-test-alpine:0.0.30 -q)
""")
    _ = wf.dispatch()

    assert wf.get_value("docker-image-id")


def test_docker_build_with_matrix(temp_file: FileIO) -> None:
    filename = str(temp_file.name)

    wf = create_workflow(f"""
jobs:
    create-docker:
        matrix:
            os: [alpine, ubuntu]

        steps:
            - name: "Build ${{{{ matrix.os }}}} Dockerfile"
              uses: core/expand-template
              with:
                  input: |
                      FROM ${{{{ matrix.os }}}}:latest
                      RUN echo "Building for ${{{{ matrix.os }}}}"
                  output_file: {filename}
            - uses: docker/build
              with:
                  dockerfile: {filename}
                  context: .
                  tags:
                      - "bluish-test-${{{{ matrix.os }}}}:test"
            - run: |
                  echo "id=$(docker image ls -f reference=bluish-test-${{{{ matrix.os }}}}:test -q)" >> "$BLUISH_OUTPUT"
              set:
                  workflow.var.docker-image-${{{{ matrix.os }}}}-id: ${{{{ outputs.id }}}}
            - run: |
                  docker image rm $(docker image ls -f reference=bluish-test-${{{{ matrix.os }}}}:test -q)
""")
    _ = wf.dispatch()

    assert wf.get_value("var.docker-image-alpine-id")
    assert wf.get_value("var.docker-image-ubuntu-id")
    assert wf.get_value("var.docker-image-alpine-id") != wf.get_value("var.docker-image-ubuntu-id")




import logging
from io import FileIO
from test.utils import create_workflow

import pytest
from bluish.core import (
    ExecutionStatus,
    init_commands,
)
from bluish.nodes import CircularDependencyError
from bluish.process import run
from bluish.schemas import RequiredAttributeError


@pytest.fixture(scope="session", autouse=True)
def initialize_commands():
    init_commands()


def test_multiple_jobs() -> None:
    wf = create_workflow("""
jobs:
    job1:
        name: "Job 1"
        steps:
            - run: echo 'This is Job 1'
    job2:
        name: "Job 2"
        steps:
            - run: echo 'This is Job 2, step 1'
            - run: echo 'This is Job 2, step 2'
""")
    _ = wf.dispatch()

    assert wf.jobs["job1"].result.stdout == "This is Job 1"
    assert wf.jobs["job2"].result.stdout == "This is Job 2, step 2"


def test_multiple_jobs_failed() -> None:
    wf = create_workflow("""
jobs:
    job1:
        name: "Job 1"
        steps:
            - run: |
                  echo 'This is Job 1'
                  false
    job2:
        name: "Job 2"
        steps:
            - run: echo 'This is Job 2, step 1'
""")
    _ = wf.dispatch()

    assert wf.jobs["job1"].result.stdout == "This is Job 1"
    assert wf.jobs["job2"].result.stdout == ""


def test_wf_matrix(temp_file: FileIO) -> None:
    filename = str(temp_file.name)

    wf = create_workflow(f"""
matrix:
    os: [ubuntu, macos]
    version: ["18.04", "20.04"]
    color: [red, blue]

jobs:
    test_job:
        name: "Job 1"
        steps:
            - run: |
                echo ${{{{ matrix.os }}}}-${{{{ matrix.version }}}}-${{{{ matrix.color }}}} >> {filename}
""")
    _ = wf.dispatch()
    output = temp_file.read().decode()

    assert output == """ubuntu-18.04-red
ubuntu-18.04-blue
ubuntu-20.04-red
ubuntu-20.04-blue
macos-18.04-red
macos-18.04-blue
macos-20.04-red
macos-20.04-blue
"""


def test_job_matrix(temp_file: FileIO) -> None:
    filename = str(temp_file.name)

    wf = create_workflow(f"""
jobs:
    test_job:
        name: "Job 1"
        matrix:
            os: [ubuntu, macos]
            version: ["18.04", "20.04"]
            color: [red, blue]
        steps:
            - run: |
                echo ${{{{ matrix.os }}}}-${{{{ matrix.version }}}}-${{{{ matrix.color }}}} >> {filename}
""")
    _ = wf.dispatch()
    output = temp_file.read().decode()

    assert output == """ubuntu-18.04-red
ubuntu-18.04-blue
ubuntu-20.04-red
ubuntu-20.04-blue
macos-18.04-red
macos-18.04-blue
macos-20.04-red
macos-20.04-blue
"""


def test_compound_matrix(temp_file: FileIO) -> None:
    filename = str(temp_file.name)

    wf = create_workflow(f"""
matrix:
    os: [ubuntu, macos]
    version: ["18.04", "20.04"]

jobs:
    test_job:
        name: "Job 1"
        matrix:
            color: [red, blue]
        steps:
            - run: |
                echo ${{{{ matrix.os }}}}-${{{{ matrix.version }}}}-${{{{ matrix.color }}}} >> {filename}
""")
    _ = wf.dispatch()
    output = temp_file.read().decode()

    assert output == """ubuntu-18.04-red
ubuntu-18.04-blue
ubuntu-20.04-red
ubuntu-20.04-blue
macos-18.04-red
macos-18.04-blue
macos-20.04-red
macos-20.04-blue
"""


def test_depends_on() -> None:
    wf = create_workflow("""
jobs:
    job1:
        name: "Job 1"
        steps:
            - run: echo 'This is Job 1'
    job2:
        name: "Job 2"
        depends_on:
            - job1
        steps:
            - run: echo 'This is Job 2, step 1'
            - run: echo 'This is Job 2, step 2'
""")
    _ = wf.dispatch_job(wf.jobs["job2"], False)
    assert wf.jobs["job1"].result.stdout == "This is Job 1"
    assert wf.jobs["job2"].result.stdout == "This is Job 2, step 2"


def test_depends_on_circular() -> None:
    wf = create_workflow("""
jobs:
    job1:
        name: "Job 1"
        depends_on:
            - job2
        steps:
            - run: echo 'This is Job 1'
    job2:
        name: "Job 2"
        depends_on:
            - job1
        steps:
            - run: echo 'This is Job 2, step 1'
            - run: echo 'This is Job 2, step 2'
""")
    try:
        _ = wf.dispatch_job(wf.jobs["job2"], False)
        raise AssertionError("Circular dependency not detected")
    except CircularDependencyError:
        pass


def test_depends_on_failed() -> None:
    wf = create_workflow("""
jobs:
    job1:
        name: "Job 1"
        steps:
            - run: |
                  echo 'This is Job 1'
                  false
    job2:
        name: "Job 2"
        depends_on:
            - job1
        steps:
            - run: echo 'This is Job 2, step 1'
            - run: echo 'This is Job 2, step 2'
""")
    _ = wf.dispatch_job(wf.jobs["job2"], False)
    assert wf.jobs["job1"].status == ExecutionStatus.FINISHED
    assert wf.jobs["job1"].result.failed
    assert wf.jobs["job2"].status == ExecutionStatus.PENDING
    assert wf.jobs["job2"].result.stdout == ""


def test_depends_on_ignored() -> None:
    wf = create_workflow("""
jobs:
    job1:
        name: "Job 1"
        steps:
            - run: echo 'This is Job 1'
    job2:
        name: "Job 2"
        depends_on:
            - job1
        steps:
            - run: echo 'This is Job 2, step 1'
            - run: echo 'This is Job 2, step 2'
""")
    _ = wf.dispatch_job(wf.jobs["job2"], True)
    assert wf.jobs["job1"].result.stdout == ""
    assert wf.jobs["job2"].result.stdout == "This is Job 2, step 2"


def test_conditions() -> None:
    wf = create_workflow("""
var:
    true_var: true
    false_var: false

jobs:
    # check == false at the job level
    job1:
        name: "Job 1"
        if: false
        steps:
            - run: echo 'This will not be printed'

    # check == true at the job level
    job2:
        name: "Job 2"
        if: ${{ true }}
        steps:
            - run: echo 'This is Job 2'

    # no check (default == true)
    job3:
        name: "Job 3"
        steps:
            - run: echo 'This is Job 3'

    # check == true at the step level
    job4:
        name: "Job 4"
        steps:
            - if: ${{ true_var }}
              shell: python
              run: |
                print("This is Job 4")

    # check == false at the step level
    job5:
        name: "Job 5"
        steps:
            - if: false_var
              shell: python
              run: |
                print("This will not be printed")
""")
    _ = wf.dispatch()

    assert wf.jobs["job1"].result.stdout == ""
    assert wf.jobs["job2"].result.stdout == "This is Job 2"
    assert wf.jobs["job3"].result.stdout == "This is Job 3"
    assert wf.jobs["job4"].result.stdout == "This is Job 4"
    assert wf.jobs["job5"].result.stdout == ""


def test_working_directory() -> None:
    wf = create_workflow("""

working_directory: /tmp

jobs:
    test_job:
        name: "Working Directory"
        steps:
            - run: pwd
""")
    _ = wf.dispatch()

    assert wf.jobs["test_job"].result.stdout == "/tmp"


def test_working_directory_override() -> None:
    wf = create_workflow("""

working_directory: /tmp

jobs:
    test_job:
        name: "Working Directory"
        working_directory: /home
        steps:
            - run: pwd
""")
    _ = wf.dispatch()

    assert wf.jobs["test_job"].result.stdout == "/home"


def test_default_run() -> None:
    wf = create_workflow("""
jobs:
    hello_run:
        name: "Hello, World!"
        steps:
            - run: echo 'Hello, World!'
""")
    _ = wf.dispatch()

    assert wf.jobs["hello_run"].result.stdout == "Hello, World!"


def test_shell_run_bash() -> None:
    wf = create_workflow("""
jobs:
    hello_sh:
        name: "Hello, World!"
        shell: bash
        steps:
            - run: echo 'Hello, World!'
""")
    _ = wf.dispatch()

    assert wf.jobs["hello_sh"].result.stdout == "Hello, World!"


def test_shell_run_python() -> None:
    wf = create_workflow("""
jobs:
    hello_python:
        name: "Hello, World!"
        steps:
            - shell: python
              run: |
                v = "World"
                print(f'Hello, {v}!')
""")
    _ = wf.dispatch()

    assert wf.jobs["hello_python"].result.stdout == "Hello, World!"


def test_shell_override() -> None:
    wf = create_workflow("""
shell: sh

jobs:
    hello_python:
        name: "Hello, World!"
        shell: node  # Overridde at job level

        steps:
            - shell: python  # Effective value
              run: |
                v = "World"
                print(f'Hello, {v}!')
""")
    _ = wf.dispatch()

    assert wf.jobs["hello_python"].result.stdout == "Hello, World!"


def test_mandatory_attributes() -> None:
    wf = create_workflow("""
jobs:
    mandatory_attributes:
        steps:
            - name: 'This step lacks a run property'
""")
    try:
        _ = wf.dispatch()
        raise AssertionError("Mandatory attribute not detected")
    except RequiredAttributeError:
        pass


def test_mandatory_inputs() -> None:
    wf = create_workflow("""
var:
    WORLD: "World!"

jobs:
    mandatory_inputs:
        steps:
            - uses: core/expand-template
              with:
                input: "Hello, ${{{{ var.WORLD }}}}"
""")
    try:
        _ = wf.dispatch()
        raise AssertionError("Mandatory input not detected")
    except RequiredAttributeError:
        pass


def test_workflow_inputs() -> None:
    wf = create_workflow("""
inputs:
    - name: TEST_INPUT
      required: true

jobs:
    test:
        steps:
            - run: |
                echo "Hello, ${{ workflow.inputs.TEST_INPUT }}!"
""")
    wf.set_inputs({"TEST_INPUT": "World"})
    _ = wf.dispatch()
    assert wf.get_value("jobs.test.stdout") == "Hello, World!"


def test_workflow_unexpected_inputs() -> None:
    try:
        wf = create_workflow("""
jobs:
    test:
        steps:
            - run: |
                echo "Hello, ${{ workflow.inputs.TEST_INPUT }}!"
""")
        wf.set_inputs({
            "UNEXPECTED_INPUT": "Unexpected"
        })
        raise AssertionError("Unexpected input not detected")
    except ValueError as ex:
        assert str(ex) == "Unknown input parameter: UNEXPECTED_INPUT"


def test_workflow_optional_inputs() -> None:
    wf = create_workflow("""
inputs:
    - name: TEST_INPUT
      required: false
      default: "World"

jobs:
    test:
        steps:
            - run: |
                echo "Hello, ${{ workflow.inputs.TEST_INPUT }}!"
""")
    wf.set_inputs({})
    _ = wf.dispatch()
    assert wf.get_value("jobs.test.stdout") == "Hello, World!"


def test_cwd() -> None:
    wf = create_workflow("""
working_directory: /tmp

jobs:
    cwd:
        steps:
            - run: pwd
""")
    _ = wf.dispatch()

    assert wf.jobs["cwd"].result.stdout == "/tmp"


def test_expansion() -> None:
    wf = create_workflow("""
var:
    HELLO: "Hello"
    WORLD: "World!"
    SMILEY: ":-)"

jobs:
    expansion:
        name: Test expansion
        steps:
            - run: echo '${{ var.HELLO }} ${{ var.WORLD }} ${{ var.SMILEY }}'
""")
    _ = wf.dispatch()

    assert wf.jobs["expansion"].result.stdout == "Hello World! :-)"


def test_expansion_with_ambiguity() -> None:
    wf = create_workflow("""
var:
    test_result: 
    stdout: "No"

jobs:
    expansion:
        name: Test expansion
        steps:
            - run: echo 'Yes!'
              set:
                workflow.var.test_result: ${{ stdout }}
""")
    try:
        _ = wf.dispatch()
        raise AssertionError("Ambiguity not detected")
    except ValueError as ex:
        assert str(ex) == "Ambiguous value reference: stdout"


def test_expansion_with_no_ambiguity() -> None:
    wf = create_workflow("""
var:
    test_result:
    stdout: "No"

jobs:
    expansion:
        name: Test expansion
        steps:
            - run: echo 'Yes!'
              set:
                workflow.var.test_result: ${{ .stdout }}
""")
    _ = wf.dispatch()
    assert wf.get_value("test_result") == "Yes!"
    assert wf.get_value("var.test_result") == "Yes!"


def test_working_dir_expansion() -> None:
    wf = create_workflow("""
var:
    PATH: "/tmp"

jobs:
    test_job:
        name: Test expansion
        var:
            DIRECTORY: ${{ var.PATH }}
        steps:
            - run: echo '${{ var.DIRECTORY }}'
""")
    _ = wf.dispatch()

    assert wf.jobs["test_job"].result.stdout == "/tmp"


def test_values() -> None:
    wf = create_workflow("""
var:
    VALUE: 1

jobs:
    test_job:
        name: Test values
        steps:
            - id: step-1
              run: echo 'VALUE == ${{ var.VALUE }}'
              set:
                  jobs.test_job.var.VALUE: 99  # Will be overridden by the next line
                  job.var.VALUE: 2
""")
    _ = wf.dispatch()

    assert wf.get_value("var.VALUE") == 1
    assert wf.get_value("workflow.var.VALUE") == 1
    assert wf.get_value("workflow.jobs.test_job.var.VALUE") == 2
    assert wf.get_value("jobs.test_job.var.VALUE") == 2
    assert wf.get_value("jobs.test_job.steps.step-1.var.VALUE") == 2
    assert wf.get_value("jobs.test_job.var.VALUE") == 2


def test_expressions() -> None:
    wf = create_workflow("""
var:
    VALUE: 2

jobs:
    test_job:
        steps:
            - run: echo 'VALUE == ${{ var.VALUE }}'
            - run: echo 'VALUE == ${{ var.VALUE + 1 }}'
            - run: echo 'VALUE == ${{ var.VALUE + 1.0 }}'
            - run: echo 'VALUE == ${{ var.VALUE + 1.5 }}'
            - run: |
                echo '${{ var.VALUE > 1 ? 'greater' : 'not greater' }}'
            - run: |
                echo '${{ var.VALUE > 3 ? 'greater' : 'not greater' }}'
            - run: |
                echo '${{ true ? 1 : 0 }}'
            - run: |
                echo '${{ 'aaa' == 'aaa' ? 1 : 0 }}'
""")
    _ = wf.dispatch()

    assert wf.get_value("jobs.test_job.steps.step_1.stdout") == "VALUE == 2"
    assert wf.get_value("jobs.test_job.steps.step_2.stdout") == "VALUE == 3"
    assert wf.get_value("jobs.test_job.steps.step_3.stdout") == "VALUE == 3.0"
    assert wf.get_value("jobs.test_job.steps.step_4.stdout") == "VALUE == 3.5"
    assert wf.get_value("jobs.test_job.steps.step_5.stdout") == "greater"
    assert wf.get_value("jobs.test_job.steps.step_6.stdout") == "not greater"
    assert wf.get_value("jobs.test_job.steps.step_7.stdout") == "1"
    assert wf.get_value("jobs.test_job.steps.step_8.stdout") == "1"


def test_secrets_are_redacted_in_log(caplog) -> None:
    caplog.set_level(logging.DEBUG)
    
    wf = create_workflow("""
secrets:
    my_secret: "hello"

var:
    my_var: "world"

jobs:
    test_job:
        name: Test values
        steps:
            - run: |
                  echo "secret is = ${{ secrets.my_secret }}"
                  echo "var is = ${{ var.my_var }}"
""")
    _ = wf.dispatch()

    assert wf.get_value("jobs.test_job.steps.step_1.stdout") == "secret is = hello\nvar is = world"
    assert "secret is = ***" in caplog.text
    assert "var is = world" in caplog.text


def test_set() -> None:
    wf = create_workflow("""
var:
    VALUE: 42

jobs:
    test_job:
        name: Test set
        steps:
            - id: step-1
              run: |
                echo 'VALUE == ${{ var.VALUE }}'
              set:
                  workflow.var.VALUE: 2
                  job.var.COOL_YEAR: 1980
                  jobs.test_job.var.OUTPUT: ${{ .stdout }}
""")
    _ = wf.dispatch()

    assert wf.get_value("var.VALUE") == 2
    assert wf.get_value("jobs.test_job.var.COOL_YEAR") == 1980
    assert wf.get_value("jobs.test_job.steps.step-1.stdout") == "VALUE == 42"
    assert wf.get_value("jobs.test_job.var.OUTPUT") == "VALUE == 42"


def test_env_overriding() -> None:
    wf = create_workflow("""
var:
    HELLO: "Hello"
    WORLD: "World!"
    SMILEY: ":-("

jobs:
    test_job:
        var:
            SMILEY: "xD"
        name: Test expansion
        steps:
            - id: step-1
              run: echo '${{ var.HELLO }} ${{ var.WORLD }} ${{ var.SMILEY }}'
              var:
                  SMILEY: ":-DDDD"
            - id: step-2
              run: echo '${{ var.HELLO }} ${{ var.WORLD }} ${{ var.SMILEY }}'

    test_job_2:
        name: Test expansion
        steps:
            - id: step-1
              run: echo '${{ var.HELLO }} ${{ var.WORLD }} ${{ var.SMILEY }}'
""")
    _ = wf.dispatch()

    assert wf.get_value("jobs.test_job.steps.step-1.stdout") == "Hello World! :-DDDD"
    assert wf.get_value("jobs.test_job.steps.step-2.stdout") == "Hello World! xD"
    assert wf.get_value("jobs.test_job_2.steps.step-1.stdout") == "Hello World! :-("


def test_expand_template(temp_file: FileIO) -> None:
    filename = str(temp_file.name)

    wf = create_workflow(f"""
var:
    hero: "Don Quijote"

jobs:
    expand_template:
        steps:
            - uses: core/expand-template
              with:
                  hide_output: true
                  input: |
                      En un lugar de la Mancha, de cuyo nombre no quiero acordarme, no ha mucho
                      tiempo que vivía un hidalgo de los de lanza en astillero, adarga antigua,
                      rocín flaco y galgo corredor. Una olla de algo más vaca que carnero,
                      salpicón las más noches, duelos y quebrantos los sábados, lantejas los
                      viernes, algún palomino de añadidura los domingos, consumían las tres
                      partes de su hacienda. El resto della concluían sayo de velarte, calzas de
                      velludo para las fiestas, con sus pantuflos de lo mesmo, y los días de
                      entresemana se honraba con su vellorí de lo más fino. Tenía en su casa una
                      ama que pasaba de los cuarenta, y una sobrina que no llegaba a los veinte,
                      y un mozo de campo y plaza, que así ensillaba el rocín como tomaba la
                      podadera. Frisaba la edad de nuestro hidalgo con los cincuenta años; era de
                      complexión recia, seco de carnes, enjuto de rostro, gran madrugador y amigo
                      de la caza. Quieren decir que tenía el sobrenombre de Quijada, o Quesada,
                      que en esto hay alguna diferencia en los autores que deste caso escriben;
                      aunque, por conjeturas verosímiles, se deja entender que se llamaba
                      Quejana. Pero esto importa poco a nuestro cuento; basta que en la narración
                      dél no se salga un punto de la verdad.
                      Hello, ${{{{ hero }}}}!
                  output_file: {filename}
    """)
    _ = wf.dispatch()
    assert run(f"tail -n1 {filename}").stdout == "Hello, Don Quijote!"
    assert wf.jobs["expand_template"].result.stdout.endswith("Hello, Don Quijote!\n")


def test_capture() -> None:
    wf = create_workflow("""
jobs:
    test_job:
        steps:
            - id: step_1
              run: |
                  echo "OUT=value 1" >> "$BLUISH_OUTPUT"
            - id: step_2
              run: |
                  echo "OUT=value 2" >> "$BLUISH_OUTPUT"
              set:
                  workflow.var.OUT: ${{ outputs.OUT }}
""")
    _ = wf.dispatch()

    assert wf.get_value("jobs.test_job.steps.step_1.outputs.OUT") == "value 1"
    assert wf.get_value("jobs.test_job.steps.step_2.outputs.OUT") == "value 2"
    assert wf.get_value("var.OUT") == "value 2"


def test_pass_env() -> None:
    wf = create_workflow("""
env:
    WORLD: "World!"

jobs:
    test_job:
        steps:
            - run: |
                  echo "Hello, $WORLD"
""")
    _ = wf.dispatch()

    assert wf.jobs["test_job"].result.stdout == "Hello, World!"


def test_runs_on_job() -> None:
    wf = create_workflow("""
jobs:
    test_job:
        runs_on: docker://alpine:3.20.3
        steps:
            - run: |
                  cat /etc/os-release
""")
    _ = wf.dispatch()

    assert "3.20.3" in wf.jobs["test_job"].result.stdout


def test_runs_on_job_automount() -> None:
    wf = create_workflow("""
jobs:
    test_job:
        runs_on:
            host: docker://alpine:3.20.3
            automount: true
        steps:
            - run: |
                  pwd
""")
    _ = wf.dispatch()

    assert wf.jobs["test_job"].result.stdout == "/mnt"


def test_runs_on_job_no_automount(temp_file) -> None:
    filename = str(temp_file.name)

    temp_file.write(b"Hello, World!")
    temp_file.flush()

    wf = create_workflow(f"""
jobs:
    test_job:
        runs_on:
            host: docker://alpine:3.20.3
            -v: {filename}:/tmp/temp-file
        steps:
            - run: |
                  cat /tmp/temp-file
""")
    _ = wf.dispatch()

    assert wf.jobs["test_job"].result.stdout == "Hello, World!"


def test_runs_on_job_automount_fails_if_mounts() -> None:
    wf = create_workflow("""
jobs:
    test_job:
        runs_on:
            host: docker://alpine:3.20.3
            automount: true
            -v: .:/mnt
        steps:
            - run: |
                  echo "nop"
""")
    try:
        _ = wf.dispatch()
        raise AssertionError("Volume mount not detected")
    except ValueError as ex:
        assert str(ex) == "To use custom volumes, set automount to false"


def test_runs_on_job_automount_fails_if_workdir() -> None:
    wf = create_workflow("""
jobs:
    test_job:
        runs_on:
            host: docker://alpine:3.20.3
            automount: true
            -w: /mnt
        steps:
            - run: |
                  echo "nop"
""")
    try:
        _ = wf.dispatch()
        raise AssertionError("Volume mount not detected")
    except ValueError as ex:
        assert str(ex) == "To use custom workdir, set automount to false"

    
def test_runs_on_workflow() -> None:
    wf = create_workflow("""
runs_on: docker://alpine:3.20.3

jobs:
    test_job:
        steps:
            - run: |
                  cat /etc/os-release
""")
    _ = wf.dispatch()

    assert "3.20.3" in wf.jobs["test_job"].result.stdout


def test_runs_on_workflow_override_in_job() -> None:
    wf = create_workflow("""
runs_on: docker://alpine:3.20.3

jobs:
    test_job:
        runs_on: docker://ubuntu:22.04
        steps:
            - run: |
                  cat /etc/os-release
""")
    _ = wf.dispatch()

    assert "22.04" in wf.jobs["test_job"].result.stdout


@pytest.mark.docker
def test_docker_pass_env() -> None:
    wf = create_workflow("""
env:
    WORLD: "World!"

jobs:
    pass_env:
        runs_on: docker://alpine:latest
        steps:
            - run: |
                  echo "Hello, $WORLD"
""")
    _ = wf.dispatch()

    assert wf.jobs["pass_env"].result.stdout == "Hello, World!"


@pytest.mark.docker
def test_docker_run() -> None:
    wf = create_workflow("""
jobs:
    docker_run:
        runs_on: docker://alpine:latest
        steps:
            - run: |
                  echo 'Hello, World!'
    """)
    _ = wf.dispatch()

    assert wf.jobs["docker_run"].result.stdout == "Hello, World!"


@pytest.mark.docker
def test_docker_file_upload(temp_file: FileIO) -> None:
    filename = str(temp_file.name)

    temp_file.write(b"Hello, World!")
    temp_file.flush()

    wf = create_workflow(f"""
jobs:
    file_upload:
        runs_on: docker://alpine:latest
        steps:
            - uses: core/upload-file
              with:
                  source_file: {filename}
                  destination_file: /tmp/hello.txt
            - run: cat /tmp/hello.txt
    """)
    _ = wf.dispatch()

    assert wf.jobs["file_upload"].result.stdout == "Hello, World!"


@pytest.mark.docker
def test_docker_file_download(temp_file: FileIO) -> None:
    filename = str(temp_file.name)
    
    temp_file.write(b"Hello, World!")
    temp_file.flush()

    wf = create_workflow(f"""
jobs:
    file_download:
        runs_on: docker://alpine:latest
        steps:
            - run: |
                  echo 'Hello, World!' > /tmp/hello.txt
            - uses: core/download-file
              with:
                  source_file: /tmp/hello.txt
                  destination_file: {filename}
    """)
    _ = wf.dispatch()

    with open(temp_file.name, "r") as f:
        file_contents = f.read().strip()
    assert file_contents == "Hello, World!"


@pytest.mark.docker
def test_docker_file_download_failed(temp_file: FileIO) -> None:
    filename = str(temp_file.name)

    temp_file.write(b"Hello, World!")
    temp_file.flush()

    wf = create_workflow(f"""
jobs:
    test_job:
        runs_on: docker://alpine:latest
        steps:
            - uses: core/download-file
              with:
                  source_file: /tmp/hello.txt
                  destination_file: {filename}
            - run: |
                  echo 'Hello, World!'
    """)
    _ = wf.dispatch()

    assert wf.jobs["test_job"].failed
    assert wf.jobs["test_job"].steps[0].failed
    assert wf.get_value("jobs.test_job.steps.step_2.stdout") == ""



import re
from typing import Any, Callable

import lark

import bluish.nodes
from bluish.safe_string import SafeString

SENSITIVE_LITERALS = ("password", "secret", "token")

EXPRESSION_REGEX = re.compile(r"\${{(.+?)}}", re.DOTALL)

EXPRESSION_GRAMMAR = r"""
    ?start: expression

    ?expression: or_expr
    ?ternary_expr: or_expr "?" or_expr ":" or_expr -> ternary

    ?or_expr: and_expr
            | or_expr "||" and_expr -> or_
            | ternary_expr

    ?and_expr: equality_expr
             | and_expr "&&" equality_expr -> and_

    ?equality_expr: relational_expr
                  | equality_expr "==" relational_expr -> eq
                  | equality_expr "!=" relational_expr -> ne

    ?relational_expr: additive_expr
                    | relational_expr "<" additive_expr  -> lt
                    | relational_expr ">" additive_expr  -> gt
                    | relational_expr "<=" additive_expr -> le
                    | relational_expr ">=" additive_expr -> ge

    ?additive_expr: multiplicative_expr
                | additive_expr "+" multiplicative_expr  -> add
                | additive_expr "-" multiplicative_expr  -> sub

    ?multiplicative_expr: unary_expr
                        | multiplicative_expr "*" unary_expr -> mul
                        | multiplicative_expr "/" unary_expr -> div
                        | multiplicative_expr "%" unary_expr -> mod

    ?unary_expr: primary_expr
               | "-" unary_expr -> neg
               | "!" unary_expr -> not_

    ?primary_expr: NUMBER             -> number
                 | VARIABLE           -> var
                 | STRING             -> str
                 | "(" expression ")"

    VARIABLE: /[a-zA-Z_.][a-zA-Z0-9_.]*/
    NUMBER: /[0-9]+(\.[0-9]+)?/
    STRING: ESCAPED_STRING | /'[^']*'/ | /"[^"]*"/

    %import common.ESCAPED_STRING
    %import common.WS

    %ignore WS
"""


_parser = lark.Lark(EXPRESSION_GRAMMAR, parser="lalr")


def to_number(value: Any) -> Any:
    if isinstance(value, (int, float)):
        return value
    elif value.isnumeric():
        return int(value)
    elif value.replace(".", "").isnumeric():
        return float(value)
    else:
        raise ValueError(f"Invalid number: {value}")


def to_bool(value: Any) -> bool:
    if isinstance(value, bool):
        return value
    elif value.isnumeric() or value.replace(".", "").isnumeric():
        return float(value) != 0
    else:
        return bool(value)


def concat(a: Any, b: Any) -> Any:
    if a is None:
        return b
    if b is None:
        return a

    result = SafeString(str(a) + str(b))
    result.redacted_value = a.redacted_value if isinstance(a, SafeString) else str(a)
    result.redacted_value += b.redacted_value if isinstance(b, SafeString) else str(b)
    return result


@lark.v_args(inline=True)
class ExprTransformer(lark.visitors.Transformer_InPlaceRecursive):
    def __init__(self, ctx: bluish.nodes.Node):
        self.expr_depth: int = 0
        self.context = ctx

    def number(self, value: str) -> int | float:
        return to_number(value)

    def str(self, value: str) -> str:
        if isinstance(value, SafeString):
            return SafeString(value[1:-1], value.redacted_value[1:-1])
        elif isinstance(value, str):
            return value[1:-1]
        else:
            return str(value)

    def expr(self, expr: Any) -> Any:
        return expr

    def add(self, a: Any, b: Any) -> Any:
        if isinstance(a, (int, float)) or isinstance(b, (int, float)):
            return a + b
        else:
            return concat(a, b)

    def sub(self, a: int | float, b: int | float) -> int | float:
        return to_number(a) - to_number(b)

    def mul(self, a: int | float, b: int | float) -> int | float:
        return to_number(a) * to_number(b)

    def div(self, a: int | float, b: int | float) -> int | float:
        return to_number(a) / to_number(b)

    def mod(self, a: int | float, b: int | float) -> int | float:
        return to_number(a) % to_number(b)

    def neg(self, a: int | float) -> int | float:
        return -to_number(a)

    def eq(self, a: Any, b: Any) -> bool:
        return a == b

    def ne(self, a: Any, b: Any) -> bool:
        return a != b

    def lt(self, a: Any, b: Any) -> bool:
        return a < b

    def gt(self, a: Any, b: Any) -> bool:
        return a > b

    def le(self, a: Any, b: Any) -> bool:
        return a <= b

    def ge(self, a: Any, b: Any) -> bool:
        return a >= b

    def and_(self, a: Any, b: Any) -> bool:
        return to_bool(a) and to_bool(b)

    def or_(self, a: Any, b: Any) -> bool:
        return to_bool(a) or to_bool(b)

    def not_(self, a: Any) -> bool:
        return not to_bool(a)

    def var(self, name) -> Any:
        if name == "true":
            return True
        elif name == "false":
            return False
        else:
            return self.context.get_value(str(name))

    def ternary(self, condition: Any, true_expr: Any, false_expr: Any) -> Any:
        return true_expr if to_bool(condition) else false_expr


def create_parser(ctx: bluish.nodes.Node) -> Callable[[str], Any]:
    """Create a parser with the given context.

    >>> ctx = context.WorkflowContext({"secrets": {"password": "1234"}})
    >>> ctx.set_value("var.x", 10)
    >>> parse = create_parser(ctx)
    >>> parse("${{ x + 10 }}")
    20.0
    >>> parse("${{ x * 10 }}")
    100.0
    >>> parse("${{ x == 10 }}")
    True
    >>> parse("${{ x != 10 }}")
    False
    >>> parse("${{ x < 10 }}")
    False
    >>> parse("${{ x > 10 }}")
    False
    >>> parse("${{ x <= 10 }}")
    True
    >>> parse("${{ x >= 10 }}")
    True
    >>> parse("${{ x && true }}")
    True
    >>> parse("${{ x || false }}")
    True
    >>> parse("${{ !x }}")
    False
    >>> bool(parse("${{ x }}"))
    True
    >>> bool(parse("${{ true }}"))
    True
    >>> bool(parse("${{ false }}"))
    False
    >>> ctx.set_value("var.x", 10)
    >>> parse("${{ x }}")
    10.0
    >>> ctx.set_value("var.x", 10.0)
    >>> parse("${{ x }}")
    10.0
    >>> ctx.set_value("var.x", "hello")
    >>> parse("${{ x }}")
    'hello'
    >>> ctx.set_value("var.x", True)
    >>> parse("${{ x }}")
    True
    >>> ctx.set_value("var.x", 1)
    >>> parse("one == ${{ x }}")
    'one == 1'
    >>> parse("one == ${{ x }}!!")
    'one == 1!!'
    >>> parse('${{ "a" + "b" }}')
    'ab'
    >>> parse('${{ "a" + secrets.password }}')
    'a1234'
    >>> parse('${{ "a" + secrets.password }}').redacted_value
    'a********'
    >>> parse('${{ secrets.password }}')
    '1234'
    """

    transformer = ExprTransformer(ctx)

    def parse(value: str) -> Any:
        result: Any = None

        offset = 0

        for m in re.finditer(EXPRESSION_REGEX, value):
            previous_chunk = value[offset : m.start()]
            ast = _parser.parse(m.group(1))
            try:
                parse_result = transformer.transform(ast)
            except lark.exceptions.VisitError as e:
                if e.orig_exc:
                    raise e.orig_exc
                else:
                    raise RuntimeError(
                        f"Error parsing expression: {m.group(1)}: {str(e)}"
                    )

            offset = m.end()

            if previous_chunk:
                result = concat(result, previous_chunk)

            result = concat(result, parse_result)

        last_chunk = value[offset:]
        if last_chunk:
            result = concat(result, last_chunk)

        return result

    return parse



# The dreaded "utils" module, where lazy programmers put all the miscellaneous functions.


from bluish.safe_string import SafeString


def safe_string(value: str | SafeString) -> str:
    """Returns the string value of a RedactedString."""
    if isinstance(value, SafeString):
        return value.redacted_value
    else:
        return value


def decorate_for_log(value: str | SafeString, decoration: str = "    ") -> str:
    """Decorates a multiline string for pretty logging."""

    def decorate(value: str, decoration: str) -> str:
        value = value.rstrip()
        if not value:
            return value

        if "\n" not in value:
            return f"{decoration}{value}"
        lines = value.rstrip().splitlines(keepends=True)
        return "\n" + "".join(f"{decoration}{line}" for line in lines)

    if isinstance(value, SafeString):
        result = SafeString(value)
        result.redacted_value = decorate(value.redacted_value, decoration)
        return result
    else:
        return decorate(value, decoration)



from typing import cast

import bluish.actions.base
import bluish.nodes.job
import bluish.nodes.step
import bluish.process
from bluish.logging import error, info
from bluish.schemas import DefaultStringList, Object, Optional, Str


class InstallPackages(bluish.actions.base.Action):
    FQN: str = "linux/install-packages"

    INPUTS_SCHEMA = Object(
        {
            "packages": DefaultStringList,
            "flavor": Optional(Str),
        }
    )

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        package_str = " ".join(step.inputs["packages"])
        flavor = step.inputs.get("flavor", "auto")

        info(f"Installing packages {package_str}...")

        job = cast(bluish.nodes.job.Job, step.parent)

        result = bluish.process.install_package(
            job.runs_on_host, step.inputs["packages"], flavor=flavor
        )
        if result.failed:
            error(f"Failed to install packages {package_str}\n{result.error}")

        return result



import contextlib
import subprocess
from typing import Any, Callable

from bluish.logging import debug, info

SHELLS = {
    "bash": "bash -euo pipefail",
    "sh": "sh -eu",
    "python": "python3 -qsIEB",
}


DEFAULT_SHELL = "sh"


class ProcessResult(subprocess.CompletedProcess[str]):
    """The result of a process execution."""

    def __init__(self, stdout: str = "", stderr: str = "", returncode: int = 0):
        super().__init__("", returncode, stdout, stderr)

    @property
    def failed(self) -> bool:
        return self.returncode != 0

    @property
    def error(self) -> str:
        return self.stderr or self.stdout

    @staticmethod
    def from_subprocess_result(
        data: subprocess.CompletedProcess[str]
    ) -> "ProcessResult":
        return ProcessResult(
            stdout=data.stdout or "",
            stderr=data.stderr or "",
            returncode=data.returncode,
        )


def _escape_command(command: str) -> str:
    return command.replace("\\", r"\\\\").replace("$", "\\$")


def _get_docker_pid(host: str, docker_args: dict[str, Any]) -> str:
    """Gets the container id from the container name or id."""

    docker_pid = run(f"docker ps -f name={host} -qa").stdout.strip()
    if docker_pid:
        info(f"Found container {host} with pid {docker_pid}")
        return docker_pid

    docker_pid = run(f"docker ps -f id={host} -qa").stdout.strip()
    if docker_pid:
        info(f"Found container {host} with pid {docker_pid}")
        return docker_pid

    info(f"Preparing container {host}...")

    opts: str = ""

    if docker_args.get("automount", False):
        if "-v" in docker_args or "--volume" in docker_args:
            raise ValueError("To use custom volumes, set automount to false")
        if "-w" in docker_args or "--workdir" in docker_args:
            raise ValueError("To use custom workdir, set automount to false")

        opts += " -v .:/mnt"
        opts += " -w /mnt"

    for k, v in docker_args.items():
        if k == "automount":
            continue
        opts += f" {k}" if isinstance(v, bool) else f" {k} {v}"

    command = f"docker run {opts} --detach {host} sleep infinity"
    debug(f" > {command}")
    run_result = run(command)
    if run_result.failed:
        raise ValueError(f"Could not start container {host}: {run_result.error}")
    docker_pid = run_result.stdout.strip()
    info(f" - Container pid {docker_pid}")

    return docker_pid


def prepare_host(opts: str | dict[str, Any] | None) -> dict[str, Any]:
    """Prepares a host for running commands."""

    host_args: dict[str, Any] | None = None
    if isinstance(opts, dict):
        host = opts.get("host", None)
        host_args = {k: v for k, v in opts.items() if k != "host"}
    else:
        host = opts

    if not host:
        return {}

    if not isinstance(host, str):
        raise ValueError(f"Invalid host: {host}")

    if host.startswith("docker://"):
        host = host[9:]
        docker_pid = _get_docker_pid(host, host_args or {})
        if not docker_pid:
            raise ValueError(f"Could not find container with name or id {host}")
        return {"host": f"docker://{docker_pid}", **(host_args if host_args else {})}
    elif host.startswith("ssh://"):
        return {"host": host, **(host_args if host_args else {})}
    else:
        raise ValueError(f"Unsupported host: {host}")


def cleanup_host(host_opts: dict[str, Any] | None) -> None:
    """Stops and removes a container if it was started by the process module."""

    if not host_opts:
        return

    host = host_opts.get("host", None) if isinstance(host_opts, dict) else host_opts
    if not host:
        return

    assert isinstance(host, str)

    if host.startswith("docker://"):
        host = host[9:]
        info(f"Stopping and removing container {host}...")

        with contextlib.suppress(Exception):
            run(f"docker stop {host}")

        with contextlib.suppress(Exception):
            run(f"docker rm {host}")


def capture_subprocess_output(
    command: str,
    stdout_handler: Callable[[str], None] | None = None,
) -> subprocess.CompletedProcess[str]:
    # Got the poll() trick from https://gist.github.com/tonykwok/e341a1413520bbb7cdba216ea7255828
    # Thanks @tonykwok!

    # shell = True is required for passing a string command instead of a list
    # bufsize = 1 means output is line buffered
    # universal_newlines = True is required for line buffering
    process = subprocess.Popen(
        command,
        shell=True,
        bufsize=1,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
        text=True,
    )

    assert process.stdout is not None
    assert process.stderr is not None

    stdout: str = ""
    stderr: str = ""

    def process_line(line: str) -> None:
        nonlocal stdout
        stdout += line
        if stdout_handler:
            stdout_handler(line.rstrip())

    while process.poll() is None:
        process_line(process.stdout.readline())

    # Process the remaining lines
    for line in process.stdout.readlines():
        process_line(line)

    return_code = process.wait()
    stdout = stdout.rstrip()
    stderr = process.stderr.read().rstrip()

    return subprocess.CompletedProcess(command, return_code, stdout, stderr)


def run(
    command: str,
    host_opts: dict[str, Any] | None = None,
    stdout_handler: Callable[[str], None] | None = None,
    stderr_handler: Callable[[str], None] | None = None,
) -> ProcessResult:
    """Runs a command on a host and returns the result.

    - `host` can be `None` (or empty) for the local host, `ssh://[user@]<host>` for an SSH host
    or `docker://<container>` for a running Docker container.
    - `stdout_handler` and `stderr_handler` are optional functions that are called
    with the output of the command as it is produced.

    Returns a `ProcessResult` object with the output of the command.
    """

    command = _escape_command(command)

    host_opts = host_opts if isinstance(host_opts, dict) else {"host": host_opts}
    host = host_opts.get("host", None)

    if host:
        if host.startswith("ssh://"):
            ssh_host = host[6:]
            opts = ""
            if "identity_file" in host_opts:
                opts += f"-i {host_opts['identity_file']}"
            command = f"ssh {opts} {ssh_host} -- '{command}'"
        elif host.startswith("docker://"):
            docker_pid = host[9:]
            command = f"docker exec -i {docker_pid} sh -euc '{command}'"

    cmd_result = capture_subprocess_output(command, stdout_handler)

    result = ProcessResult.from_subprocess_result(cmd_result)
    if result.failed and result.stderr and stderr_handler:
        stderr_handler(cmd_result.stderr)

    return result


def get_flavor(host_opts: dict[str, Any] | None) -> str:
    ids = {}
    for line in run("cat /etc/os-release | grep ^ID", host_opts).stdout.splitlines():
        key, value = line.split("=", maxsplit=1)
        ids[key] = value.strip().strip('"')
    return ids.get("ID_LIKE", ids.get("ID", "Unknown"))


def install_package(
    host_opts: dict[str, Any] | None, packages: list[str], flavor: str = "auto"
) -> ProcessResult:
    """Installs a package on a host."""

    package_list = " ".join(packages)
    if not packages:
        raise ValueError("Empty package list")

    flavor = get_flavor(host_opts) if flavor == "auto" else flavor
    if flavor in ("alpine", "alpine-edge"):
        return run(f"apk update && apk add {package_list}", host_opts)
    elif flavor in ("debian", "ubuntu"):
        return run(f"apt-get update && apt-get install -y {package_list}", host_opts)
    elif flavor in ("fedora", "centos", "rhel", "rocky"):
        return run(f"dnf install -y {package_list}", host_opts)
    elif flavor in ("arch"):
        return run(f"pacman -S --noconfirm {package_list}", host_opts)
    elif flavor in ("suse", "opensuse", "opensuse-leap", "opensuse-tumbleweed"):
        return run(f"zypper install -y {package_list}", host_opts)
    elif flavor in ("gentoo"):
        return run(f"emerge -v {package_list}", host_opts)
    elif flavor in ("macos"):
        brew_install_cmd = (
            "HOMEBREW_NO_AUTO_UPDATE=1 HOMEBREW_NO_INSTALL_CLEANUP=1 brew install"
        )

        package_list = " ".join(p for p in packages if not p.startswith("cask:"))
        if package_list:
            result = run(f"{brew_install_cmd} {package_list}", host_opts)

        cask_list = " ".join(p[5:] for p in packages if not p.startswith("cask:"))
        if cask_list:
            result = run(f"{brew_install_cmd} --cask {cask_list}", host_opts)

        return result
    else:
        raise ValueError(f"Unsupported flavor: {flavor}")



import contextlib
import logging
import os
from io import StringIO
from typing import Never

import click
import yaml
from flask import Flask, abort, jsonify, request

from bluish.__main__ import PROJECT_VERSION
from bluish.core import (
    init_commands,
)
from bluish.nodes import WorkflowDefinition
from bluish.nodes.job import Job
from bluish.nodes.workflow import Workflow


class LogFormatter(logging.Formatter):
    COLORS = {
        logging.DEBUG: "white",
        logging.INFO: "bright_white",
        logging.WARNING: "yellow",
        logging.ERROR: "red",
        logging.CRITICAL: "bright_red",
    }

    def __init__(self, format: str) -> None:
        super().__init__(fmt=format)

    def format(self, record: logging.LogRecord) -> str:
        record.msg = click.style(
            record.msg, fg=self.COLORS.get(record.levelno, "white")
        )
        return super().format(record)


def fatal(message: str, exit_code: int = 1) -> Never:
    click.secho(message, fg="red", bold=True)
    exit(exit_code)


def init_logging(level_name: str) -> None:
    for level in [
        logging.INFO,
        logging.DEBUG,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]:
        logging.addLevelName(level, "")

    log_level = getattr(logging, level_name.upper(), logging.INFO)
    logging.basicConfig(level=log_level)
    logging.getLogger().handlers[0].setFormatter(
        LogFormatter("%(levelname)s%(message)s")
    )


def locate_yaml(name: str) -> str | None:
    """Locates the workflow file."""

    if not name:
        name = "bluish"

    paths = (name, f".bluish/{name}")
    exts = (".yaml", ".yml")

    for path in paths:
        for ext in exts:
            if os.path.exists(f"{path}{ext}"):
                return f"{path}{ext}"

    return None


def workflow_from_file(file: str) -> Workflow:
    """Loads the workflow from a file."""

    yaml_contents: str = ""

    with contextlib.suppress(FileNotFoundError):
        with open(file, "r") as yaml_file:
            yaml_contents = yaml_file.read()

    if not yaml_contents:
        fatal("No workflow file found.")

    definition = WorkflowDefinition(**yaml.safe_load(yaml_contents))
    return Workflow(definition)


@click.command("blu")
@click.argument("job_id", type=str, required=True)
@click.option("--no-deps", is_flag=True, help="Don't run job dependencies")
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"]),
    default="INFO",
    help="Log level",
)
@click.version_option(PROJECT_VERSION)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def blu_cli(
    job_id: str,
    no_deps: bool,
    log_level: str,
    args: tuple[str],
) -> None:
    init_logging(log_level)
    init_commands()

    file: str = ""
    if ":" in job_id:
        file, job_id = job_id.split(":")

    yaml_path = locate_yaml(file)
    if not yaml_path:
        fatal("No workflow file found.")

    logging.info(f"Loading workflow from {yaml_path}")
    logging.info("")

    wf = workflow_from_file(yaml_path)
    wf.set_inputs({k: v for k, v in (arg.split("=", maxsplit=1) for arg in args)})

    job: Job | None = wf.jobs.get(job_id)
    if not job:
        fatal(f"Job '{job_id}' not found.")

    try:
        result = wf.dispatch_job(job, no_deps)
        if not result:
            return
        if result.failed:
            exit(result.returncode)
        else:
            click.secho("Job completed successfully.", fg="green")

    except Exception as e:
        if os.environ.get("BLUISH_DEBUG"):
            import traceback

            trace = traceback.format_exc()
            print(trace)
        fatal(str(e))


@click.group("bluish")
@click.option(
    "--file", "-f", type=click.Path(dir_okay=False, readable=True, resolve_path=True)
)
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"]),
    default="INFO",
    help="Log level",
)
@click.version_option(PROJECT_VERSION)
@click.pass_context
def bluish_cli(
    ctx: click.Context,
    file: str,
    log_level: str,
) -> None:
    init_logging(log_level)
    init_commands()

    yaml_contents: str = ""
    yaml_path = file or locate_yaml("")
    if not yaml_path:
        fatal("No workflow file found.")

    logging.info(f"Loading workflow from {yaml_path}")
    logging.info("")

    with contextlib.suppress(FileNotFoundError):
        with open(yaml_path, "r") as yaml_file:
            yaml_contents = yaml_file.read()

    if not yaml_contents:
        fatal("No workflow file found.")

    definition = WorkflowDefinition(**yaml.safe_load(yaml_contents))
    wf = Workflow(definition)
    ctx.obj = wf


@bluish_cli.command("list")
@click.pass_obj
def list_jobs(wf: Workflow) -> None:
    if not wf.jobs:
        fatal("No jobs found in workflow file.")

    items = tuple((id, job.attrs.name or "") for id, job in wf.jobs.items())
    id_len = max(len(id) for id, _ in items)

    click.secho(f"{'ID':<{id_len}}  NAME", fg="yellow", bold=True)
    for id, name in items:
        click.echo(f"{id:<{id_len}}  {name}")


@bluish_cli.command("run")
@click.argument("job_id", type=str, required=True)
@click.option("--no-deps", is_flag=True, help="Don't run job dependencies")
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.pass_obj
def run_job(wf: Workflow, job_id: str, no_deps: bool, args: tuple[str]) -> None:
    wf.set_inputs({k: v for k, v in (arg.split("=", maxsplit=1) for arg in args)})
    job = wf.jobs.get(job_id)
    if not job:
        fatal(f"Job '{job_id}' not found.")

    try:
        result = wf.dispatch_job(job, no_deps)
        if not result:
            return
        elif result.failed:
            exit(result.returncode)
        else:
            click.secho("Job completed successfully.", fg="green")
    except Exception as e:
        if os.environ.get("BLUISH_DEBUG"):
            import traceback

            trace = traceback.format_exc()
            print(trace)

        fatal(str(e))


@bluish_cli.command("serve")
@click.argument(
    "workflow_path",
    type=click.Path(file_okay=False, dir_okay=True, exists=True),
    required=True,
)
@click.option("--host", type=str, default="localhost", help="Host")
@click.option("--port", type=int, default=5000, help="Port")
def serve(workflow_path: str, host: str, port: int) -> None:
    app = Flask(__name__)

    @app.route("/")
    def serve_index():
        return abort(404)

    @app.route("/<file>/<job_id>")
    def dispatch_job(file: str, job_id: str):
        file = f"{file}.yaml"
        if file not in os.listdir(workflow_path):
            return abort(404)

        wf = workflow_from_file(f"{workflow_path}/{file}")
        job = wf.jobs.get(job_id)
        if not job:
            return abort(404)

        log_level_name = request.args.get("log_level", "INFO")
        log_level = getattr(logging, log_level_name.upper(), logging.INFO)

        log_stream = StringIO()
        logging.basicConfig(
            stream=log_stream, level=log_level, format="%(message)s", force=True
        )

        result = wf.dispatch_job(job, False)

        return jsonify(
            {
                "run": result is not None,
                "stdout": log_stream.getvalue().splitlines(),
                "stderr": result.stderr if result else None,
                "returncode": result.returncode if result else None,
            }
        )

    app.run(host=host, port=port)


if __name__ == "__main__":
    pass



import os
from typing import Any

from dotenv import dotenv_values

import bluish.core
import bluish.nodes
import bluish.nodes.job
import bluish.process
from bluish.logging import debug, error, info


class Workflow(bluish.nodes.Node):
    NODE_TYPE = "workflow"

    def __init__(self, definition: bluish.nodes.Definition) -> None:
        super().__init__(None, definition)

        self.matrix: dict
        self.sys_env: dict
        self.jobs: dict
        self.runs_on_host: dict[str, Any] | None
        self._inputs: dict[str, str] | None = None

        self._job_definitions: dict = {}
        for k, v in self.attrs.jobs.items():
            v["id"] = k
            self._job_definitions[k] = bluish.nodes.JobDefinition(**v)

        self.reset()

    @property
    def inputs(self) -> dict[str, str]:
        return self._inputs or {}

    def reset(self) -> None:
        self.matrix = {}
        self.sys_env = {}
        self.jobs = {}
        self.runs_on_host = None

        self.secrets.update(
            {
                k: v
                for k, v in dotenv_values(self.attrs.secrets_file or ".secrets").items()
                if v is not None
            }
        )

        self.sys_env = {
            **os.environ,
            **dotenv_values(self.attrs.env_file or ".env"),
        }

        for k, v in self._job_definitions.items():
            self.jobs[k] = bluish.nodes.job.Job(self, v)

    def set_inputs(self, inputs: dict[str, str]) -> None:
        def is_true(v: Any) -> bool:
            return v in ("true", "1", True)

        self._inputs = {}
        for param in self.attrs.inputs:
            name = param.get("name")
            if not name:
                raise ValueError("Invalid input parameter (missing name)")

            if is_true(param.get("sensitive")):
                self.sensitive_inputs.add(name)

            if name in inputs or "default" in param:
                self._inputs[name] = self.expand_expr(
                    inputs.get(name, param.get("default"))
                )
            elif is_true(param.get("required")):
                raise ValueError(f"Missing required input parameter: {name}")

        # Check for unknown input parameters
        unknowns = list(k for k in inputs.keys() if k not in self._inputs)
        if unknowns:
            if len(unknowns) == 1:
                raise ValueError(f"Unknown input parameter: {unknowns[0]}")
            else:
                raise ValueError(f"Unknown input parameters: {unknowns}")

    def dispatch(self) -> bluish.process.ProcessResult:
        self.reset()

        self.status = bluish.core.ExecutionStatus.RUNNING

        bluish.nodes.log_dict(
            self.inputs,
            header="with",
            ctx=self,
            sensitive_keys=self.sensitive_inputs,
        )

        try:
            for job in self.jobs.values():
                result = self.dispatch_job(job, no_deps=False)
                if not result:
                    continue

                self.result = result
                if result.failed and not job.attrs.continue_on_error:
                    self.failed = True
                    break

            return self.result

        finally:
            if self.status == bluish.core.ExecutionStatus.RUNNING:
                self.status = bluish.core.ExecutionStatus.FINISHED

    def dispatch_job(
        self, job: bluish.nodes.job.Job, no_deps: bool
    ) -> bluish.process.ProcessResult | None:
        return self.__dispatch_job(job, no_deps, set())

    def __dispatch_job(
        self, job: bluish.nodes.job.Job, no_deps: bool, visited_jobs: set[str]
    ) -> bluish.process.ProcessResult | None:
        if job.id in visited_jobs:
            raise bluish.nodes.CircularDependencyError("Circular reference detected")

        if job.status == bluish.core.ExecutionStatus.FINISHED:
            info(f"Job {job.id} already dispatched and finished")
            return job.result
        elif job.status == bluish.core.ExecutionStatus.SKIPPED:
            info(f"Re-running skipped job {job.id}")

        visited_jobs.add(job.id)

        if not no_deps:
            debug("Getting dependency map...")
            for dependency_id in job.attrs.depends_on or []:
                dep_job = self.jobs.get(dependency_id)
                if not dep_job:
                    raise RuntimeError(f"Invalid dependency job id: {dependency_id}")

                result = self.__dispatch_job(dep_job, no_deps, visited_jobs)
                if result and result.failed:
                    error(f"Dependency {dependency_id} failed")
                    return result

        for wf_matrix in bluish.nodes._generate_matrices(self):

            self.matrix = wf_matrix

            cleanup_host = False
            if not self.runs_on_host:
                self.runs_on_host = self.runs_on_host or bluish.process.prepare_host(
                    self.expand_expr(self.attrs.runs_on)
                )
                cleanup_host = True

            try:
                for job_matrix in bluish.nodes._generate_matrices(job):
                    job.reset()

                    job.matrix = {**wf_matrix, **job_matrix}

                    if job.attrs.runs_on:
                        job.runs_on_host = bluish.process.prepare_host(
                            self.expand_expr(job.attrs.runs_on)
                        )
                    else:
                        job.runs_on_host = self.runs_on_host

                
                    try:
                        result = job.dispatch()
                        if result and result.failed:
                            return result
                    except:
                        raise
                    finally:
                        if job.runs_on_host and job.runs_on_host is not self.runs_on_host:
                            bluish.process.cleanup_host(job.runs_on_host)
                            job.runs_on_host = None
            except:
                raise
            finally:
                if cleanup_host:
                    bluish.process.cleanup_host(self.runs_on_host)
                    self.runs_on_host = None

        return bluish.process.ProcessResult()



import os
from collections import ChainMap
from typing import cast

import bluish.actions.base
import bluish.nodes.job
import bluish.nodes.step
import bluish.process
from bluish.logging import error, info
from bluish.schemas import Int, List, Object, Optional, Str


class RunCommand(bluish.actions.base.Action):
    FQN: str = ""
    SCHEMA = Object(
        {
            "run": Str,
        }
    )

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        env = ChainMap(step.env, step.parent.env, step.parent.parent.env)  # type: ignore

        bluish.nodes.log_dict(
            env, header="env", ctx=step, sensitive_keys=self.SENSITIVE_INPUTS
        )

        if not step.attrs.run:
            return bluish.process.ProcessResult()

        command = step.attrs.run.strip()

        echo_commands = step.get_inherited_attr("echo_commands", True)
        echo_output = step.get_inherited_attr("echo_output", True)
        assert echo_commands is not None
        assert echo_output is not None

        command = step.expand_expr(command)

        if echo_commands:
            info(command)

        return step.parent.exec(command, step, env=env, stream_output=echo_output)  # type: ignore


class ExpandTemplate(bluish.actions.base.Action):
    FQN: str = "core/expand-template"

    INPUTS_SCHEMA = Object(
        {
            "input": Optional(Str, List(Str)),
            "input_file": Optional(Str),
            "output_file": Str,
            "chmod": Optional(Int, Str),
        }
    )

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        inputs = step.inputs

        job = cast(bluish.nodes.job.Job, step.parent)

        template_content: str
        if "input_file" in inputs:
            template_file = inputs["input_file"]

            info(f"Reading template file: {template_file}...")
            template_content = job.read_file(template_file).decode()
        else:
            template_content = inputs["input"]

        expanded_content = step.expand_expr(template_content)

        output_file = inputs.get("output_file")
        if output_file:
            info(f"Writing expanded content to: {output_file}...")
            job.write_file(output_file, expanded_content.encode())

            if "chmod" in inputs:
                permissions = inputs["chmod"]
                info(f"Setting permissions to {permissions} on {output_file}...")
                _ = job.exec(f"chmod {permissions} {output_file}", step)

        return bluish.process.ProcessResult(stdout=expanded_content)


class UploadFile(bluish.actions.base.Action):
    FQN: str = "core/upload-file"

    INPUTS_SCHEMA = Object(
        {
            "source_file": Str,
            "destination_file": Str,
            "chmod": Optional(Int, Str),
        }
    )

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        inputs = step.inputs

        contents: str

        source_file = inputs["source_file"]
        source_file = os.path.expanduser(source_file)

        info(f"Reading file: {source_file}...")
        with open(source_file, "r") as f:
            contents = f.read()
        info(f" - Read {len(contents)} bytes.")

        destination_file = inputs.get("destination_file")
        assert destination_file is not None

        job = cast(bluish.nodes.job.Job, step.parent)

        info(f"Writing file to: {destination_file}...")
        try:
            job.write_file(destination_file, contents.encode())
            result = bluish.process.ProcessResult()
        except IOError as e:
            error(f"Failed to write file: {str(e)}")
            return bluish.process.ProcessResult(returncode=1)

        if "chmod" in inputs:
            permissions = inputs["chmod"]
            info(f"Setting permissions to {permissions} on {destination_file}...")
            chmod_result = job.exec(f"chmod {permissions} {destination_file}", step)
            if chmod_result.failed:
                error(f"Failed to set permissions: {result.stderr}")
                return chmod_result

        return result


class DownloadFile(bluish.actions.base.Action):
    FQN: str = "core/download-file"

    INPUTS_SCHEMA = Object(
        {
            "source_file": Str,
            "destination_file": Str,
            "chmod": Optional(Int, Str),
        }
    )

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        inputs = step.inputs

        source_file = inputs["source_file"]
        info(f"Reading file: {source_file}...")
        try:
            job = cast(bluish.nodes.job.Job, step.parent)
            raw_contents = job.read_file(source_file)
        except IOError as e:
            error(f"Failed to read file: {str(e)}")
            return bluish.process.ProcessResult(returncode=1)

        info(f" - Read {len(raw_contents)} bytes.")

        destination_file: str | None = None

        try:
            destination_file = inputs.get("destination_file")
            if destination_file:
                info(f"Writing file to: {destination_file}...")
                with open(destination_file, "wb") as f:
                    f.write(raw_contents)
        except Exception as e:
            error(f"Failed to write file: {str(e)}")
            return bluish.process.ProcessResult(returncode=1)

        if destination_file and "chmod" in inputs:
            permissions = inputs["chmod"]
            info(f"Setting permissions to {permissions} on {destination_file}...")
            try:
                os.chmod(destination_file, permissions)
            except Exception as e:
                error(f"Failed to set permissions: {str(e)}")
                return bluish.process.ProcessResult(returncode=1)

        return bluish.process.ProcessResult()




from test.utils import create_workflow

import pytest
from bluish.core import init_commands


@pytest.fixture(scope="session", autouse=True)
def initialize_commands():
    init_commands()


@pytest.mark.docker
def test_install_packages_ubuntu() -> None:
    wf = create_workflow("""
jobs:
  test_job:
    runs_on: docker://ubuntu:latest
    steps:
      - uses: linux/install-packages
        with:
          packages:
            - moreutils
            - jq
""")
    _ = wf.dispatch()
    assert wf.jobs["test_job"].result.failed is False


@pytest.mark.docker
def test_install_packages_alpine() -> None:
    wf = create_workflow("""
jobs:
  test_job:
    runs_on: docker://alpine:latest
    steps:
      - uses: linux/install-packages
        with:
          packages:
            - moreutils
            - jq
""")
    _ = wf.dispatch()
    assert wf.jobs["test_job"].result.failed is False


@pytest.mark.docker
def test_install_packages_failed() -> None:
    wf = create_workflow("""
jobs:
  test_job:
    runs_on: docker://alpine:latest
    steps:
      - uses: linux/install-packages
        with:
          packages:
              - thispackagedoesnotexist-1.0.0
""")
    _ = wf.dispatch()
    assert wf.jobs["test_job"].result.failed



import bluish.actions
import bluish.core
import bluish.nodes as nodes
import bluish.process
from bluish.logging import info


class Step(nodes.Node):
    NODE_TYPE = "step"

    def __init__(self, parent: nodes.Node, definition: nodes.Definition):
        super().__init__(parent, definition)

        self.action_class = bluish.actions.get_action(self.attrs.uses)
        if self.action_class is None:
            raise ValueError(f"Unknown action: {self.attrs.uses}")

    @property
    def display_name(self) -> str:
        if self.attrs.name:
            return self.attrs.name
        elif self.action_class.FQN:  # type: ignore
            return self.action_class.FQN  # type: ignore
        elif self.attrs.run:
            return self.attrs.run.split("\n", maxsplit=1)[0]
        else:
            return self.id

    def dispatch(self) -> bluish.process.ProcessResult | None:
        info(f"* Run step '{self.display_name}'")

        if not nodes.can_dispatch(self):
            self.status = bluish.core.ExecutionStatus.SKIPPED
            info(" >>> Skipped")
            return None

        self.status = bluish.core.ExecutionStatus.RUNNING

        try:
            assert self.action_class is not None

            if self.attrs.uses:
                info(f"Running {self.attrs.uses}")

            self.result = self.action_class().execute(self)
            self.failed = self.result.failed

        finally:
            self.status = bluish.core.ExecutionStatus.FINISHED

        return self.result



from enum import Enum


class ExecutionStatus(Enum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    FINISHED = "FINISHED"
    SKIPPED = "SKIPPED"


def init_commands() -> None:
    import bluish.actions.core  # noqa
    import bluish.actions.linux  # noqa
    import bluish.actions.macos  # noqa
    import bluish.actions.docker  # noqa
    import bluish.actions.git  # noqa

    actions = [
        bluish.actions.core.RunCommand,
        bluish.actions.core.ExpandTemplate,
        bluish.actions.core.DownloadFile,
        bluish.actions.core.UploadFile,
        bluish.actions.linux.InstallPackages,
        bluish.actions.macos.InstallPackages,
        bluish.actions.docker.Build,
        bluish.actions.docker.Run,
        bluish.actions.docker.Login,
        bluish.actions.docker.Logout,
        bluish.actions.docker.GetPid,
        bluish.actions.docker.CreateNetwork,
        bluish.actions.docker.Exec,
        bluish.actions.docker.Stop,
        bluish.actions.git.Checkout,
    ]

    from bluish.actions import register_action

    for klass in actions:
        register_action(klass)



import logging
from typing import Dict

_REGISTERED_ACTIONS: Dict[str, type] = {}


def get_action(fqn: str) -> type | None:
    return _REGISTERED_ACTIONS.get(fqn)


def register_action(klass: type) -> None:
    if not hasattr(klass, "FQN"):
        raise ValueError("Action class must have an FQN attribute")
    if klass.FQN in _REGISTERED_ACTIONS and _REGISTERED_ACTIONS[klass.FQN] != klass:
        raise ValueError(f"Action {klass.FQN} is already registered")
    if not klass.FQN:
        logging.debug(f"Registering default action from {klass.__name__}")
    else:
        logging.debug(f"Registering action {klass.FQN} from {klass.__name__}")
    _REGISTERED_ACTIONS[klass.FQN] = klass



import base64
import re
from collections import ChainMap, namedtuple
from itertools import product
from typing import Any, Callable, Generator, Iterable, Optional, Sequence, TypeVar, cast

import bluish.core
import bluish.process
from bluish.logging import info
from bluish.safe_string import SafeString
from bluish.schemas import (
    JOB_SCHEMA,
    STEP_SCHEMA,
    WORKFLOW_SCHEMA,
    Validator,
)
from bluish.utils import safe_string

TResult = TypeVar("TResult")


def log_dict(
    _dict: dict | ChainMap,
    header: str,
    ctx: "Node | None" = None,
    sensitive_keys: Sequence[str] | Iterable[str] = (),
) -> None:
    if not _dict:
        return
    info(f"{header}:")
    for k, v in _dict.items():
        if ctx:
            v = ctx.expand_expr(v)
        if k in sensitive_keys:
            info(f"  {k}: ********")
        else:
            info(f"  {k}: {safe_string(v)}")


class Definition:
    SCHEMA: Validator | None = None

    def __init__(self, **kwargs: Any):
        self.__dict__["_attrs"] = kwargs
        self._validate_attrs(kwargs)

    def as_dict(self) -> dict[str, Any]:
        return self.__dict__["_attrs"]

    def get(self, name: str, default: Any = None) -> Any:
        return self.__dict__["_attrs"].get(name, default)

    def _validate_attrs(self, attrs: dict[str, Any]):
        if self.SCHEMA:
            self.SCHEMA.validate(attrs)

    def __getattr__(self, name: str) -> Any:
        if name == "attrs":
            return self.__dict__["_attrs"]
        if name.startswith("_"):
            name = name[1:]
        return self.__dict__["_attrs"].get(name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            name = name[1:]
        self.__dict__["_attrs"][name] = value

    def __contains__(self, name: str) -> bool:
        if name.startswith("_"):
            name = name[1:]
        return name in self.__dict__["_attrs"]


class WorkflowDefinition(Definition):
    SCHEMA = WORKFLOW_SCHEMA
    pass


class JobDefinition(Definition):
    SCHEMA = JOB_SCHEMA
    pass


class StepDefinition(Definition):
    SCHEMA = STEP_SCHEMA
    pass


class Node:
    NODE_TYPE: str = ""

    def __init__(self, parent: Optional["Node"], definition: Definition):
        self._parent = parent
        self.attrs = definition

        self.result = bluish.process.ProcessResult()
        self.failed = False
        self.status = bluish.core.ExecutionStatus.PENDING
        self.sensitive_inputs: set[str] = {"password", "token"}

        self._expression_parser: Callable[[str], Any] | None = None

    @property
    def parent(self) -> "Node":
        return self._parent  # type: ignore

    def __contains__(self, name: str) -> bool:
        return name in self.__dict__ or name in self.attrs

    def __getattr__(self, name: str) -> Any:
        return getattr(self.attrs, name)

    @property
    def id(self) -> str:
        return self.attrs.id

    @property
    def name(self) -> str:
        return self.attrs.name

    @property
    def env(self) -> dict[str, Any]:
        return self.attrs.env

    @property
    def var(self) -> dict[str, Any]:
        return self.attrs.var

    @property
    def secrets(self) -> dict[str, str]:
        return self.attrs.secrets

    @property
    def inputs(self) -> dict[str, Any]:
        return self.attrs._with

    @property
    def outputs(self) -> dict[str, Any]:
        return self.attrs.outputs

    @property
    def display_name(self) -> str:
        return self.attrs.name if self.attrs.name else self.id

    @property
    def expression_parser(self) -> Callable[[str], Any]:
        # HACK This doesn't make me happy
        from bluish.expressions import create_parser

        if not self._expression_parser:
            self._expression_parser = create_parser(self)
        return self._expression_parser

    def dispatch(self) -> bluish.process.ProcessResult | None:
        raise NotImplementedError()

    def expand_expr(self, value: Any) -> Any:
        if isinstance(value, str):
            return _expand_expr(self, value)
        elif isinstance(value, list):
            return [_expand_expr(self, v) for v in value]
        elif isinstance(value, dict):
            return {k: _expand_expr(self, v) for k, v in value.items()}
        else:
            return value

    def get_value(self, name: str, default: Any = None, raw: bool = False) -> Any:
        value = _try_get_value(self, name, raw=raw)
        if value is None and default is None:
            raise ValueError(f"Variable {name} not found")
        return value if value is not None else default

    def set_value(self, name: str, value: Any) -> None:
        if not _try_set_value(self, name, value):
            raise ValueError(f"Invalid variable name: {name}")

    def set_attr(self, name: str, value: Any) -> None:
        setattr(self.attrs, name, value)

    def get_inherited_attr(
        self, name: str, default: TResult | None = None
    ) -> TResult | None:
        result = default
        ctx: Node | None = self
        while ctx is not None:
            if name in ctx:
                result = cast(TResult, getattr(ctx, name))
                break
            elif name in ctx.attrs:
                result = cast(TResult, getattr(ctx.attrs, name))
                break
            else:
                ctx = ctx.parent
        return self.expand_expr(result)


class CircularDependencyError(Exception):
    pass


class VariableExpandError(Exception):
    pass


EXPR_REGEX = re.compile(r"\$?\$\{\{\s*([a-zA-Z_.][a-zA-Z0-9_.-]*)\s*\}\}")


ValueResult = namedtuple("ValueResult", ["value", "contains_secrets"])


def _step_or_job(ctx: Node) -> Node:
    if ctx.NODE_TYPE == "step" or ctx.NODE_TYPE == "job":
        return ctx
    raise ValueError(f"Can't find step or job in context of type: {ctx.NODE_TYPE}")


def _step(ctx: Node) -> Node:
    if ctx.NODE_TYPE == "step":
        return ctx
    raise ValueError(f"Can't find step in context of type: {ctx.NODE_TYPE}")


def _job(ctx: Node) -> Node:
    if ctx.NODE_TYPE == "job":
        return ctx
    elif ctx.NODE_TYPE == "step":
        return ctx.parent  # type: ignore
    raise ValueError(f"Can't find job in context of type: {ctx.NODE_TYPE}")


def _workflow(ctx: Node) -> Node:
    if ctx.NODE_TYPE == "workflow":
        return ctx
    elif ctx.NODE_TYPE == "job":
        return ctx.parent  # type: ignore
    elif ctx.NODE_TYPE == "step":
        return ctx.parent.parent  # type: ignore
    raise ValueError(f"Can't find workflow in context of type: {ctx.NODE_TYPE}")


def _generate_matrices(ctx: Node) -> Generator[dict[str, Any], None, None]:
    if not ctx.attrs.matrix:
        yield {}
        return

    for matrix_tuple in product(*ctx.attrs.matrix.values()):
        yield {
            key: ctx.expand_expr(value)
            for key, value in zip(ctx.attrs.matrix.keys(), matrix_tuple)
        }


def _try_get_value(ctx: Node, name: str, raw: bool = False) -> Any:
    import bluish.nodes.job
    import bluish.nodes.step
    import bluish.nodes.workflow

    def prepare_value(value: Any) -> Any:
        if value is None:
            return None
        elif raw or not isinstance(value, str):
            return value
        else:
            return cast(str, _expand_expr(ctx, value))

    if "." not in name:
        # Handle a non-fully qualified variable name and avoid ambiguity
        member_result = _try_get_value(ctx, f".{name}", raw=raw)
        var_result = _try_get_value(ctx, f"var.{name}", raw=raw)

        if var_result is not None and member_result is not None:
            raise ValueError(f"Ambiguous value reference: {name}")
        elif var_result is not None:
            return var_result
        elif member_result is not None:
            return member_result
        else:
            return None

    root, varname = name.split(".", maxsplit=1)

    if root == "":
        if varname == "stdout":
            return prepare_value(
                "" if ctx.result is None else ctx.result.stdout.strip()
            )
        elif varname == "stderr":
            return prepare_value(
                "" if ctx.result is None else ctx.result.stderr.strip()
            )
        elif varname == "returncode":
            return prepare_value(0 if ctx.result is None else ctx.result.returncode)
    elif root == "env":
        current: Node | None = ctx
        while current:
            for k in ("sys_env", "env"):
                dict_ = getattr(current, k, None)
                if dict_ and varname in dict_:
                    return prepare_value(dict_[varname])
            current = current.parent
    elif root == "var":
        current = ctx
        while current:
            if varname in current.var:
                return prepare_value(current.var[varname])
            current = current.parent
    elif root == "workflow":
        return _try_get_value(_workflow(ctx), varname, raw)
    elif root == "secrets":
        wf = cast(bluish.nodes.workflow.Workflow, _workflow(ctx))
        if varname in wf.secrets:
            return prepare_value(SafeString(cast(str, wf.secrets[varname]), "********"))
    elif root == "jobs":
        wf = cast(bluish.nodes.workflow.Workflow, _workflow(ctx))
        job_id, varname = varname.split(".", maxsplit=1)
        job = wf.jobs.get(job_id)
        if not job:
            raise ValueError(f"Job {job_id} not found")
        return _try_get_value(job, varname, raw)
    elif root == "job":
        return _try_get_value(_job(ctx), varname, raw)
    elif root == "steps":
        job = cast(bluish.nodes.job.Job, _job(ctx))
        step_id, varname = varname.split(".", maxsplit=1)
        step = next((step for step in job.steps if step.id == step_id), None)
        if not step:
            raise ValueError(f"Step {step_id} not found")
        return _try_get_value(step, varname, raw)
    elif root == "matrix":
        matrix = getattr(ctx, "matrix", None) or getattr(ctx.parent, "matrix", None)
        if matrix and (varname in matrix):
            return prepare_value(matrix[varname])
    elif root == "step":
        return _try_get_value(_step(ctx), varname, raw)
    elif root == "inputs":
        node = ctx
        if varname in node.inputs:
            if varname in node.sensitive_inputs:
                return prepare_value(SafeString(node.inputs[varname], "********"))
            else:
                return prepare_value(node.inputs[varname])
    elif root == "outputs":
        node = _step_or_job(ctx)
        if varname in node.outputs:
            return prepare_value(node.outputs[varname])

    return None


def _try_set_value(ctx: "Node", name: str, value: str) -> bool:
    import bluish.nodes.job
    import bluish.nodes.step
    import bluish.nodes.workflow

    if "." not in name:
        return False

    name = cast(str, _expand_expr(ctx, name))
    root, varname = name.split(".", maxsplit=1)
    if root == "":
        root, varname = varname.split(".", maxsplit=1)

    if root == "env":
        ctx.env[varname] = value
        return True
    elif root == "var":
        ctx.var[varname] = value
        return True
    elif root == "workflow":
        return _try_set_value(_workflow(ctx), varname, value)
    elif root == "jobs":
        wf = cast(bluish.nodes.workflow.Workflow, _workflow(ctx))
        job_id, varname = varname.split(".", maxsplit=1)
        job = wf.jobs.get(job_id)
        if not job:
            raise ValueError(f"Job {job_id} not found")
        return _try_set_value(job, varname, value)
    elif root == "job":
        return _try_set_value(_job(ctx), varname, value)
    elif root == "steps":
        job = cast(bluish.nodes.job.Job, _job(ctx))
        step_id, varname = varname.split(".", maxsplit=1)
        step = next((step for step in job.steps if step.id == step_id), None)
        if not step:
            raise ValueError(f"Step {step_id} not found")
        return _try_set_value(step, varname, value)
    elif root == "step":
        return _try_set_value(_step(ctx), varname, value)
    elif root == "inputs":
        step = cast(bluish.nodes.step.Step, _step(ctx))
        step.inputs[varname] = value
        return True
    elif root == "outputs":
        node = _step_or_job(ctx)
        node.outputs[varname] = value
        return True

    return False


TExpandValue = str | dict[str, Any] | list[str]


def _expand_expr(
    ctx: Node, value: TExpandValue | None, _depth: int = 1
) -> TExpandValue:
    if not isinstance(value, str):
        if isinstance(value, dict):
            return {k: _expand_expr(ctx, v, _depth=_depth) for k, v in value.items()}
        elif isinstance(value, list):
            return [cast(str, _expand_expr(ctx, v, _depth=_depth)) for v in value]
        else:
            return value  # type: ignore

    if "${{" not in value:
        return value

    return ctx.expression_parser(value)


def can_dispatch(context: Node) -> bool:
    if context.attrs._if is None:
        return True

    info(f"Testing {context.attrs._if}")
    if isinstance(context.attrs._if, bool):
        return context.attrs._if
    elif not isinstance(context.attrs._if, str):
        raise ValueError("Condition must be a bool or a string")

    # Allow bare `if` expressions without placeholders
    condition = context.attrs._if
    if "${{" not in condition:
        condition = "${{" + condition + "}}"

    return bool(context.expand_expr(condition))


def _read_file(ctx: Node, file_path: str) -> bytes:
    """Reads a file from a host and returns its content as bytes."""

    import bluish.nodes.job

    job = cast(bluish.nodes.job.Job, _job(ctx))
    result = job.exec(f"base64 -i '{file_path}'", ctx)
    if result.failed:
        raise IOError(f"Failure reading from {file_path}: {result.error}")

    return base64.b64decode(result.stdout)


def _write_file(ctx: Node, file_path: str, content: bytes) -> None:
    """Writes content to a file on a host."""

    import bluish.nodes.job

    job = cast(bluish.nodes.job.Job, _job(ctx))
    b64 = base64.b64encode(content).decode()

    result = job.exec(f"echo {b64} | base64 -di - > {file_path}", ctx)
    if result.failed:
        raise IOError(f"Failure writing to {file_path}: {result.error}")



import inspect
from typing import Any as TAny
from typing import Iterable, Union, cast

import bluish.process
from bluish.safe_string import SafeString


class Validator:
    def __init__(self, *types: type | TAny | None, **kwargs: TAny) -> None:
        self._types = tuple(_ensure_validator_instance(t) for t in types)
        self._args = kwargs
        self._default_value = self._get_default_raw_value()

    @property
    def has_default_value(self) -> bool:
        return self._default_value is not None

    def get_default_value(self) -> TAny | None:
        if self._default_value is not None:
            return (
                self._default_value()
                if callable(self._default_value)
                else self._default_value
            )
        return None

    def _get_default_raw_value(self) -> TAny | None:
        _default: TAny | None = None

        if "default" not in self._args:
            for t in self._types:
                if t is None or not isinstance(t, Validator):
                    continue

                _default = t._get_default_raw_value()
                if _default is not None:
                    return _default
            return None
        else:
            return self._args["default"]

    def validate(self, data: TAny) -> None:
        for t in self._types:
            if _validate_type(t, data):
                return
        raise InvalidTypeError(self, None, data)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._types})"


TTypeValidator = Union[type, Validator, None]


class InvalidTypeError(Exception):
    def __init__(
        self, type: TTypeValidator, property: str | None, value: TAny | None
    ) -> None:
        super().__init__()
        self.type = type
        self.property = property
        self.value = value

    def __str__(self) -> str:
        if self.property is None:
            return f"{self.value} is not any of the allowed types: {self.type}"
        return f"Property '{self.property}' with value {self.value} is not any of the allowed types: {self.type}"


class RequiredAttributeError(Exception):
    def __init__(self, attr: str):
        super().__init__()
        self.attr = attr

    def __str__(self) -> str:
        return f"Required attribute '{self.attr}' is missing"


class UnexpectedAttributesError(Exception):
    def __init__(self, attrs: Iterable[str]):
        super().__init__(f"Unexpected attributes: {attrs}")


def _validate_type(t: TTypeValidator, data: TAny) -> bool:
    if t is None:
        return True
    elif t is TAny:
        return True
    elif isinstance(t, Validator):
        t.validate(data)
        return True
    elif isinstance(data, cast(type, t)):
        return True
    else:
        return False


def _validate_or_fail(t: TTypeValidator, data: TAny) -> None:
    if not _validate_type(t, data):
        raise InvalidTypeError(t, None, data)


def _ensure_validator_instance(t: TTypeValidator) -> TTypeValidator:
    """
    Ensure that if we receive an uninstanciated Validator class, we return an instance of it.
    """
    if t is None:
        return None
    elif inspect.isclass(t) and issubclass(t, Validator):
        return t()
    else:
        return t


class Str(Validator):
    def __init__(self, **kwargs: TAny) -> None:
        super().__init__(str, SafeString, **kwargs)

    def __repr__(self) -> str:
        return "string"


class Int(Validator):
    def __init__(self, **kwargs: TAny) -> None:
        super().__init__(int, **kwargs)

    def __repr__(self) -> str:
        return "int"


class Float(Validator):
    def __init__(self, **kwargs: TAny) -> None:
        super().__init__(float, int, **kwargs)

    def __repr__(self) -> str:
        return "float"


class Bool(Validator):
    def __init__(self, **kwargs: TAny) -> None:
        super().__init__(bool, **kwargs)

    def __repr__(self) -> str:
        return "bool"


class AnyType(Validator):
    def __init__(self, **kwargs: TAny) -> None:
        super().__init__(TAny, **kwargs)

    def __repr__(self) -> str:
        return "any"


class Dict(Validator):
    def __init__(
        self,
        key_schema: TTypeValidator = Str,
        value_schema: TTypeValidator = AnyType,
        **kwargs: TAny,
    ) -> None:
        super().__init__(dict, **kwargs)
        self._key_schema = _ensure_validator_instance(key_schema)
        self._value_schema = _ensure_validator_instance(value_schema)

    def validate(self, data: TAny) -> None:
        if not isinstance(data, dict):
            raise InvalidTypeError(self, None, data)

        for k, v in data.items():
            _validate_or_fail(self._key_schema, k)
            _validate_or_fail(self._value_schema, v)

    def __repr__(self) -> str:
        return f"dict<{repr(self._key_schema)}, {repr(self._value_schema)}>"


class Object(Validator):
    def __init__(self, properties: dict[str, TTypeValidator], **kwargs: TAny) -> None:
        reject_extra = kwargs.pop("reject_extra", True)

        super().__init__(dict, **kwargs)
        self._properties = {
            k: _ensure_validator_instance(v) for k, v in properties.items()
        }
        self._reject_extra = reject_extra

    def validate(self, data: TAny) -> None:
        if not isinstance(data, dict):
            raise InvalidTypeError(self, None, data)

        all_props = self._properties.items()
        required = {k for k, t in all_props if not isinstance(t, Optional)}
        optional = {k for k, t in all_props if isinstance(t, Optional)}

        def ensure_property(k: str) -> bool:
            t = self._properties[k]
            if k not in data and isinstance(t, Validator) and t.has_default_value:
                data[k] = t.get_default_value()
                return True
            return False

        def validate_property(k: str) -> None:
            try:
                _validate_or_fail(self._properties[k], data[k])
            except InvalidTypeError:
                raise InvalidTypeError(self._properties[k], k, data[k])

        for k in required:
            if k not in data:
                if not ensure_property(k):
                    raise RequiredAttributeError(k)
                continue
            try:
                validate_property(k)
            except InvalidTypeError as ex:
                raise InvalidTypeError(self._properties[k], k, data[k]) from ex
            except RequiredAttributeError as ex:
                raise RequiredAttributeError(f"{k}.{ex.attr}") from ex

        for k in optional:
            if k not in data:
                _ = ensure_property(k)
                continue
            try:
                validate_property(k)
            except InvalidTypeError as ex:
                raise InvalidTypeError(self._properties[k], k, data[k]) from ex
            except RequiredAttributeError as ex:
                raise RequiredAttributeError(f"{k}.{ex.attr}") from ex

    def __repr__(self) -> str:
        return f"object<{', '.join(f'{k}: {v}' for k, v in self._properties.items())}>"


class List(Validator):
    def __init__(self, item_schema: type | Validator, **kwargs: TAny) -> None:
        super().__init__(list, **kwargs)
        self._item_schema = _ensure_validator_instance(item_schema)

    def validate(self, data: TAny) -> None:
        if not isinstance(data, list):
            raise InvalidTypeError(self, None, data)

        for item in data:
            _validate_or_fail(self._item_schema, item)

    def __repr__(self) -> str:
        return f"list<{repr(self._item_schema)}>"


class Optional(Validator):
    def __init__(self, *types: type | Validator, **kwargs: TAny) -> None:
        super().__init__(None, *types, **kwargs)

    def __repr__(self) -> str:
        return f"Optional<{', '.join(repr(t) for t in self._types)}>"


DefaultDict = Dict(default=dict)

DefaultStringDict = Dict(Str, Str, default=dict)

DefaultStringList = List(Str, default=list)

_COMMON_PROPERTIES = {
    "id": Optional(Str),
    "name": Optional(Str),
    "description": Optional(Str),
    "env": DefaultDict,
    "var": DefaultDict,
    "secrets": DefaultStringDict,
    "secrets_file": Optional(Str),
    "env_file": Optional(Str),
    "if": Optional(Str, Bool),
    "echo_commands": Bool(default=True),
    "echo_output": Bool(default=True),
    "continue_on_error": Bool(default=False),
    "with": DefaultDict,
    "outputs": DefaultDict,
    "set": DefaultDict,
}

STEP_SCHEMA = Object(
    {
        **_COMMON_PROPERTIES,
        "uses": Str(default=""),
        "shell": Str(default=bluish.process.DEFAULT_SHELL),
    }
)


INPUT_DEFINITION_SCHEMA = Object(
    {
        "name": Str,
        "description": Optional(Str),
        "required": Bool(default=False),
        "sensitive": Bool(default=False),
        "default": Optional(AnyType),
    }
)


JOB_SCHEMA = Object(
    {
        **_COMMON_PROPERTIES,
        "inputs": List(INPUT_DEFINITION_SCHEMA, default=list),
        "runs_on": Optional(Str),
        "depends_on": DefaultStringList,
        "steps": List(STEP_SCHEMA),
    }
)


WORKFLOW_SCHEMA = Object(
    {
        **_COMMON_PROPERTIES,
        "inputs": List(INPUT_DEFINITION_SCHEMA, default=list),
        "runs_on": Optional(Str),
        "jobs": Dict(Str, JOB_SCHEMA),
    }
)



import os
from typing import cast

import bluish.actions.base
import bluish.nodes.job
import bluish.nodes.step
import bluish.process
from bluish.logging import error, info
from bluish.schemas import Int, Object, Optional, Str
from bluish.utils import safe_string


def run_git_command(
    command: str, step: bluish.nodes.step.Step
) -> bluish.process.ProcessResult:
    preamble: str = ""

    key_file = step.inputs.get("ssh_key_file")
    if key_file:
        preamble = f"export GIT_SSH_COMMAND='ssh -i {key_file} -o IdentitiesOnly=yes -o StrictHostKeychecking=no';"

    job = cast(bluish.nodes.job.Job, step.parent)
    return job.exec(f"{preamble} {command}", step)


def prepare_environment(step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
    REQUIRED = {
        "git": "git",
        "openssh-client": "ssh",
    }

    job = cast(bluish.nodes.job.Job, step.parent)

    required_packages = [
        package
        for package, binary in REQUIRED.items()
        if job.exec(f"which {binary}", step).failed
    ]
    if required_packages:
        info(f"Installing missing packages: {required_packages}...")
        result = bluish.process.install_package(job.runs_on_host, required_packages)
        if result.failed:
            error(f"Failed to install required packages. Error: {result.error}")
            return result

    return bluish.process.ProcessResult()


def cleanup_environment(step: bluish.nodes.step.Step) -> None:
    pass


class Checkout(bluish.actions.base.Action):
    FQN: str = "git/checkout"

    INPUTS_SCHEMA = Object(
        {
            "repository": Str,
            "depth": Optional(Int),
            "branch": Optional(Str),
        }
    )

    SENSITIVE_INPUTS: tuple[str, ...] = ("ssh_key_file", "password")

    @property
    def fqn(self) -> str:
        return "git/checkout"

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        try:
            inputs = step.inputs

            repository: str = step.expand_expr(inputs["repository"])
            repo_name = os.path.basename(repository)

            result = prepare_environment(step)
            if result.failed:
                return result

            options = ""
            if "depth" in inputs:
                options += f"--depth {inputs['depth']}"
            else:
                options += "--depth 1"

            if "branch" in inputs:
                options += f" --branch {inputs['branch']}"

            info(f"Cloning repository: {safe_string(repository)}...")
            clone_result = run_git_command(
                f"git clone {repository} {options} ./{repo_name}", step
            )
            if clone_result.failed:
                error(f"Failed to clone repository: {clone_result.error}")
                return clone_result

            # Update the current job working dir to the newly cloned repo
            info(f"Setting working directory to: {repo_name}...")
            wd = step.get_inherited_attr("working_directory", ".")
            step.parent.set_attr("working_directory", f"{wd}/{repo_name}")  # type: ignore

            return clone_result
        finally:
            cleanup_environment(step)



from typing import cast


class SafeString(str):
    """A string with two values: one for logging and one for the actual value."""

    def __new__(cls, value: str, redacted_value: str | None = None) -> "SafeString":
        result = cast(SafeString, super().__new__(cls, value))
        result.redacted_value = redacted_value or value
        return result

    @property
    def redacted_value(self) -> str:
        return getattr(self, "_redacted_value", self)

    @redacted_value.setter
    def redacted_value(self, value: str) -> None:
        self._redacted_value = value



from logging import critical as logging_critical
from logging import debug as logging_debug
from logging import error as logging_error
from logging import exception as logging_exception
from logging import info as logging_info
from logging import log as logging_log
from logging import warning as logging_warning
from typing import Any

from bluish.safe_string import SafeString


def info(message: str, *args: Any, **kwargs: Any) -> None:
    msg = message.redacted_value if isinstance(message, SafeString) else message
    logging_info(msg, *args, **kwargs)


def error(message: str, *args: Any, **kwargs: Any) -> None:
    msg = message.redacted_value if isinstance(message, SafeString) else message
    logging_error(msg, *args, **kwargs)


def warning(message: str, *args: Any, **kwargs: Any) -> None:
    msg = message.redacted_value if isinstance(message, SafeString) else message
    logging_warning(msg, *args, **kwargs)


def debug(message: str, *args: Any, **kwargs: Any) -> None:
    msg = message.redacted_value if isinstance(message, SafeString) else message
    logging_debug(msg, *args, **kwargs)


def critical(message: str, *args: Any, **kwargs: Any) -> None:
    msg = message.redacted_value if isinstance(message, SafeString) else message
    logging_critical(msg, *args, **kwargs)


def exception(message: str, *args: Any, **kwargs: Any) -> None:
    msg = message.redacted_value if isinstance(message, SafeString) else message
    logging_exception(msg, *args, **kwargs)


def log(level: int, message: str, *args: Any, **kwargs: Any) -> None:
    msg = message.redacted_value if isinstance(message, SafeString) else message
    logging_log(level, msg, *args, **kwargs)




from test.utils import create_workflow

import pytest
from bluish.core import init_commands


@pytest.fixture(scope="session", autouse=True)
def initialize_commands():
    init_commands()


def test_docker_public_checkout() -> None:
    wf = create_workflow("""
jobs:
  checkout:
    runs_on: docker://alpine:latest
    steps:
      - uses: git/checkout
        with:
          repository: https://github.com/luismedel/bluish.git
      - run: |
          head -n1 README.md
""")
    _ = wf.dispatch()
    assert wf.jobs["checkout"].result.stdout == "[![justforfunnoreally.dev badge](https://img.shields.io/badge/justforfunnoreally-dev-9ff)](https://justforfunnoreally.dev)"




import base64
from typing import Any
from uuid import uuid4

import bluish.core
import bluish.nodes
import bluish.process
from bluish.logging import debug, error, info, warning
from bluish.utils import decorate_for_log


class Job(bluish.nodes.Node):
    NODE_TYPE = "job"

    def __init__(
        self,
        parent: bluish.nodes.Node,
        definition: bluish.nodes.Definition,
    ):
        import bluish.nodes.step

        super().__init__(parent, definition)

        self.runs_on_host: dict[str, Any] | None = None
        self.matrix: dict[str, Any]
        self.steps: list[bluish.nodes.step.Step]

        self._step_definitions: list[bluish.nodes.StepDefinition] = []
        for i, step_dict in enumerate(self.attrs.steps):
            step_dict["id"] = step_dict.get("id", f"step_{i+1}")
            self._step_definitions.append(bluish.nodes.StepDefinition(**step_dict))

        self.reset()

    def reset(self) -> None:
        import bluish.nodes.step

        self._runs_on_host = None
        self.matrix = {}
        self.steps = []

        for step_def in self._step_definitions:
            step = bluish.nodes.step.Step(self, step_def)
            self.steps.append(step)

    def dispatch(self) -> bluish.process.ProcessResult | None:
        self.status = bluish.core.ExecutionStatus.RUNNING

        info(f"** Run job '{self.display_name}'")

        try:
            bluish.nodes.log_dict(self.matrix, header="matrix", ctx=self)

            if not bluish.nodes.can_dispatch(self):
                self.status = bluish.core.ExecutionStatus.SKIPPED
                info("Job skipped")
                return None

            for step in self.steps:
                result = step.dispatch()
                if not result:
                    continue

                self.result = result

                if result.failed and not step.attrs.continue_on_error:
                    self.failed = True
                    break

        finally:
            if self.status == bluish.core.ExecutionStatus.RUNNING:
                self.status = bluish.core.ExecutionStatus.FINISHED
            if self._runs_on_host:
                bluish.process.cleanup_host(self._runs_on_host)
            self._runs_on_host = None

        return self.result

    def read_file(self, file_path: str) -> bytes:
        return bluish.nodes._read_file(self, file_path)

    def write_file(self, file_path: str, content: bytes) -> None:
        bluish.nodes._write_file(self, file_path, content)

    def exec(
        self,
        command: str,
        context: bluish.nodes.Node,
        env: dict[str, str] | None = None,
        shell: str | None = None,
        stream_output: bool = False,
    ) -> bluish.process.ProcessResult:
        command = context.expand_expr(command).strip()
        if "${{" in command:
            raise ValueError("Command contains unexpanded variables")

        if context.get_inherited_attr("is_sensitive", False):
            stream_output = False

        # Define where to capture the output with the >> operator
        capture_filename = f"/tmp/{uuid4().hex}"
        debug(f"Capture file: {capture_filename}")
        touch_result = bluish.process.run(
            f"touch {capture_filename}", self.runs_on_host
        )
        if touch_result.failed:
            error(
                f"Failed to create capture file {capture_filename}: {touch_result.error}"
            )
            return touch_result

        env = env or {}

        if "BLUISH_OUTPUT" in env:
            warning(
                "BLUISH_OUTPUT is a reserved environment variable. Overwriting it.",
            )

        env["BLUISH_OUTPUT"] = capture_filename

        env_str = "; ".join([f'{k}="{v}"' for k, v in env.items()]).strip()
        if env_str:
            command = f"{env_str}; {command}"

        if shell is None:
            shell = context.get_inherited_attr("shell", bluish.process.DEFAULT_SHELL)
        assert shell is not None

        interpreter = bluish.process.SHELLS.get(shell, shell)
        if interpreter:
            b64 = base64.b64encode(command.encode()).decode()
            command = f"echo {b64} | base64 -di - | {interpreter}"

        working_dir = context.get_inherited_attr("working_directory")
        if working_dir:
            debug(f"Working dir: {working_dir}")
            debug("Making sure working directory exists...")
            mkdir_result = bluish.process.run(
                f"mkdir -p {working_dir}", self.runs_on_host
            )
            if mkdir_result.failed:
                error(
                    f"Failed to create working directory {working_dir}: {mkdir_result.error}"
                )
                return mkdir_result

            command = f'cd "{working_dir}" && {command}'

        def stdout_handler(line: str) -> None:
            info(decorate_for_log(line.rstrip(), "  > "))

        def stderr_handler(line: str) -> None:
            error(decorate_for_log(line.rstrip(), " ** "))

        run_result = bluish.process.run(
            command,
            host_opts=self.runs_on_host,
            stdout_handler=stdout_handler if stream_output else None,
            stderr_handler=stderr_handler if stream_output else None,
        )

        # HACK: We should use our own process.read_file here,
        # but it currently causes an infinite recursion
        output_result = bluish.process.run(f"cat {capture_filename}", self.runs_on_host)
        if output_result.failed:
            error(f"Failed to read capture file: {output_result.error}")
            return output_result

        for line in output_result.stdout.splitlines():
            k, v = line.split("=", maxsplit=1)
            context.outputs[k] = v

        return run_result




from bluish.schemas import (
    AnyType,
    DefaultDict,
    Dict,
    InvalidTypeError,
    List,
    Object,
    Optional,
    RequiredAttributeError,
    Str,
)


def test_happy() -> None:
    schema = Object({
        "name": Str,
        "env": DefaultDict,
    })

    schema.validate({
        "name": "test",
        "env": {"a": "b"},
    })


def test_default_values() -> None:
    schema = Object({
        "name": Str,
        "env": DefaultDict,
        "items": List(Str, default=["a", "b"]),
    })

    data = {
        "name": "test",
    }
    # This should add the default values for "env" and "items"
    schema.validate(data)

    assert "env" in data
    assert data["env"] == {}

    assert "items" in data
    assert data["items"] == ["a", "b"]


def test_missing_key() -> None:
    schema = Object({
        "name": Str,
        "env": Dict,
    })

    try:
        schema.validate({
            "name": "test",
        })
        raise AssertionError("Should have raised an exception")
    except RequiredAttributeError:
        pass


def test_optional_key() -> None:
    schema = Object({
        "name": Optional(Str),
    })

    schema.validate({})
    schema.validate({
        "name": "test",
    })

def test_lists() -> None:
    schema = Object({
        "values": List(Str),
    })
    
    schema.validate({
        "values": ["a", "b"],
    })


def test_lists_any() -> None:
    schema = Object({
        "values": List(AnyType),
    })

    schema.validate({
        "values": ["a", "b", 1, {"a": "b"}, [1, 2, 3]],
    })


def test_lists_ko() -> None:
    schema = Object({
        "values": List(Str),
    })

    try:
        schema.validate({
            "values": ["a", "b", 1],
        })
        raise AssertionError("Should have raised an exception")
    except InvalidTypeError:
        pass



from typing import Sequence

import bluish.nodes.step
import bluish.process
from bluish.logging import debug
from bluish.nodes import Definition, log_dict
from bluish.schemas import Validator


def _key_exists(key: str, attrs: Definition) -> bool:
    """Checks if a key (or pipe-separated alternative keys) exists in a dictionary."""
    return ("|" in key and any(i in attrs for i in key.split("|"))) or key in attrs


class Action:
    FQN: str = ""

    SCHEMA: Validator | None = None
    INPUTS_SCHEMA: Validator | None = None
    SENSITIVE_INPUTS: Sequence[str] = tuple()

    def run(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        raise NotImplementedError()

    def execute(self, step: bluish.nodes.step.Step) -> bluish.process.ProcessResult:
        if self.SCHEMA:
            self.SCHEMA.validate(step.attrs.as_dict())
        if self.INPUTS_SCHEMA and step.inputs:
            self.INPUTS_SCHEMA.validate(step.inputs)

        step.sensitive_inputs.update(self.SENSITIVE_INPUTS)

        log_dict(
            step.inputs,
            header="with",
            ctx=step,
            sensitive_keys=self.SENSITIVE_INPUTS,
        )

        step.result = self.run(step)

        if step.attrs.set:
            variables = step.attrs.set
            for key, value in variables.items():
                value = step.expand_expr(value)
                debug(f"Setting {key} = {value}")
                step.set_value(key, value)

        return step.result
